from builtins import object

from common.models import School
from django import forms
from django.core.exceptions import ObjectDoesNotExist
from django.core.validators import EmailValidator
from django_countries.widgets import CountrySelectWidget


class OrganisationForm(forms.ModelForm):
    class Meta(object):

        model = School
        fields = ["name", "postcode", "country"]
        widgets = {
            "name": forms.TextInput(attrs={"autocomplete": "off", "placeholder": "Name of school or club"}),
            "postcode": forms.TextInput(attrs={"autocomplete": "off", "placeholder": "Postcode / Zipcode"}),
            "country": CountrySelectWidget(layout="{widget}"),
        }
        help_texts = {"name": "Name of school or club", "postcode": "Postcode / Zipcode", "country": "Country"}

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop("user", None)
        self.current_school = kwargs.pop("current_school", None)
        super(OrganisationForm, self).__init__(*args, **kwargs)
        self.fields["postcode"].strip = False

    def clean(self):
        name = self.cleaned_data.get("name", None)
        postcode = self.cleaned_data.get("postcode", None)

        if name and postcode:
            try:
                school = School.objects.get(name=name, postcode=postcode)
            except ObjectDoesNotExist:
                return self.cleaned_data

            if not self.current_school or self.current_school.id != school.id:
                raise forms.ValidationError("There is already a school or club registered with that name and postcode")

        return self.cleaned_data

    def clean_name(self):
        name = self.cleaned_data.get("name", None)
        validator = EmailValidator()

        if name:
            try:
                validator(name)
                is_email = True
            except forms.ValidationError:
                is_email = False

            if is_email:
                raise forms.ValidationError("Please make sure your organisation name is valid")

        return name

    def clean_postcode(self):
        postcode = self.cleaned_data.get("postcode", None)

        if postcode:
            if len(postcode.replace(" ", "")) > 10 or len(postcode.replace(" ", "")) == 0:
                raise forms.ValidationError("Please enter a valid postcode or ZIP code")

        return postcode



from __future__ import absolute_import

import datetime

import time
from common.models import JoinReleaseStudent
from common.tests.utils import email as email_utils
from common.tests.utils.classes import create_class_directly
from common.tests.utils.organisation import create_organisation_directly, join_teacher_to_organisation
from common.tests.utils.student import (
    create_independent_student,
    create_independent_student_directly,
    create_school_student_directly,
    signup_duplicate_independent_student_fail,
    verify_email,
)
from common.tests.utils.teacher import signup_teacher_directly
from django.contrib.auth.models import User
from django.core import mail
from django.test import Client, TestCase
from django.urls import reverse
from selenium.webdriver.support.wait import WebDriverWait

from portal.forms.error_messages import INVALID_LOGIN_MESSAGE
from .base_test import BaseTest
from .pageObjects.portal.home_page import HomePage
from .utils.messages import (
    is_email_verified_message_showing,
    is_indep_student_join_request_received_message_showing,
    is_indep_student_join_request_revoked_message_showing,
    is_student_details_updated_message_showing,
    is_email_updated_message_showing,
    is_password_updated_message_showing,
)


class TestIndependentStudent(TestCase):
    def test_signup_short_password_fails(self):
        c = Client()

        response = c.post(
            reverse("register"),
            {
                "independent_student_signup-date_of_birth_day": 7,
                "independent_student_signup-date_of_birth_month": 10,
                "independent_student_signup-date_of_birth_year": 1997,
                "independent_student_signup-name": "Test Name",
                "independent_student_signup-email": "test@email.com",
                "independent_student_signup-consent_ticked": "on",
                "independent_student_signup-password": "pass",
                "independent_student_signup-confirm_password": "pass",
                "g-recaptcha-response": "something",
            },
        )

        # Assert response isn't a redirect (submit failure)
        assert response.status_code == 200

    def test_signup_common_password_fails(self):
        c = Client()

        response = c.post(
            reverse("register"),
            {
                "independent_student_signup-date_of_birth_day": 7,
                "independent_student_signup-date_of_birth_month": 10,
                "independent_student_signup-date_of_birth_year": 1997,
                "independent_student_signup-name": "Test Name",
                "independent_student_signup-email": "test@email.com",
                "independent_student_signup-consent_ticked": "on",
                "independent_student_signup-password": "Password1",
                "independent_student_signup-confirm_password": "Password1",
                "g-recaptcha-response": "something",
            },
        )

        # Assert response isn't a redirect (submit failure)
        assert response.status_code == 200

    def test_signup_passwords_do_not_match_fails(self):
        c = Client()

        response = c.post(
            reverse("register"),
            {
                "independent_student_signup-date_of_birth_day": 7,
                "independent_student_signup-date_of_birth_month": 10,
                "independent_student_signup-date_of_birth_year": 1997,
                "independent_student_signup-name": "Test Name",
                "independent_student_signup-email": "test@email.com",
                "independent_student_signup-consent_ticked": "on",
                "independent_student_signup-password": "Password1!",
                "independent_student_signup-confirm_password": "Password2!",
                "g-recaptcha-response": "something",
            },
        )

        # Assert response isn't a redirect (submit failure)
        assert response.status_code == 200

    def test_signup_invalid_name_fails(self):
        c = Client()

        response = c.post(
            reverse("register"),
            {
                "independent_student_signup-date_of_birth_day": 7,
                "independent_student_signup-date_of_birth_month": 10,
                "independent_student_signup-date_of_birth_year": 1997,
                "independent_student_signup-name": "///",
                "independent_student_signup-email": "test@email.com",
                "independent_student_signup-consent_ticked": "on",
                "independent_student_signup-password": "Password1!",
                "independent_student_signup-confirm_password": "Password1!",
                "g-recaptcha-response": "something",
            },
        )

        # Assert response isn't a redirect (submit failure)
        assert response.status_code == 200

    def test_signup_under_13_sends_parent_email(self):
        c = Client()

        response = c.post(
            reverse("register"),
            {
                "independent_student_signup-date_of_birth_day": datetime.date.today().day,
                "independent_student_signup-date_of_birth_month": datetime.date.today().month,
                "independent_student_signup-date_of_birth_year": datetime.date.today().year,
                "independent_student_signup-name": "Young person",
                "independent_student_signup-email": "test@email.com",
                "independent_student_signup-consent_ticked": "on",
                "independent_student_signup-password": "Password1!",
                "independent_student_signup-confirm_password": "Password1!",
                "g-recaptcha-response": "something",
            },
        )

        assert response.status_code == 302
        assert len(mail.outbox) == 1
        assert mail.outbox[0].subject == "Code for Life account request"


# Class for Selenium tests. We plan to replace these and turn them into Cypress tests
class TestIndependentStudentFrontend(BaseTest):
    def test_delete_indy_account(self):
        page = self.go_to_homepage()
        page, _, _, email, password = create_independent_student(page)
        page = page.independent_student_login(email, password)
        page = page.go_to_account_page()

        # save the user to check if it was anonymised
        user = User.objects.get(email=email)
        user_id = user.id

        # first check if a wrong password triggers the error
        unsubscribe_newsletter_checkbox = page.browser.find_element_by_name("unsubscribe_newsletter")
        unsubscribe_newsletter_checkbox.click()

        delete_account_form = page.browser.find_element_by_name("delete_password")
        delete_account_form.send_keys("123")  # wrong password

        delete_account_button = page.browser.find_element_by_id("delete_account_button")
        delete_account_button.click()
        assert (
            page.browser.find_element_by_css_selector("#form-delete-indy-account > ul > li").text
            == "Incorrect password"
        )

        # now delete the account
        unsubscribe_newsletter_checkbox = page.browser.find_element_by_name("unsubscribe_newsletter")
        unsubscribe_newsletter_checkbox.click()

        delete_account_form = page.browser.find_element_by_name("delete_password")
        delete_account_form.send_keys(password)

        delete_account_button = page.browser.find_element_by_id("delete_account_button")
        delete_account_button.click()

        delete_button_confirm = page.browser.find_element_by_id("delete_button")
        delete_button_confirm.click()

        # check if can still login to the account
        page = self.go_to_homepage()
        assert page.go_to_independent_student_login_page().independent_student_login_failure(email, password)

        # now check if anonymised
        assert not User.objects.get(id=user_id).is_active

        # check if email has been sent
        assert len(mail.outbox) == 1

    def test_signup_without_newsletter(self):
        page = self.go_to_homepage()
        page, _, _, _, _ = create_independent_student(page)
        assert is_email_verified_message_showing(self.selenium)

    def test_signup_duplicate_email_failure(self):
        page = self.go_to_homepage()
        page, _, _, email, _ = create_independent_student(page)
        assert is_email_verified_message_showing(self.selenium)

        page = self.go_to_homepage()
        page, _, _, _, _ = signup_duplicate_independent_student_fail(page, duplicate_email=email)

        assert len(mail.outbox) == 1
        assert mail.outbox[0].subject == "Duplicate account"

        assert self.is_login_page(page)

    def test_signup_duplicate_email_with_teacher(self):
        teacher_email, _ = signup_teacher_directly()

        page = self.go_to_homepage()
        page = page.go_to_signup_page()

        page = page.independent_student_signup(
            "indy", teacher_email, password="Password1!", confirm_password="Password1!"
        )

        page.return_to_home_page()

        assert len(mail.outbox) == 1
        assert mail.outbox[0].subject == "Duplicate account"

    def test_login_failure(self):
        page = self.go_to_homepage()
        page = page.go_to_independent_student_login_page()
        page = page.independent_student_login_failure("non-existent-email@codeforlife.com", "Incorrect password")

        assert page.has_login_failed("independent_student_login_form", INVALID_LOGIN_MESSAGE)

    def test_login_success(self):
        page = self.go_to_homepage()
        page, _, username, _, password = create_independent_student(page)
        page = page.independent_student_login(username, password)
        assert self.is_dashboard(page)

    def test_login_not_verified(self):
        username, password, _ = create_independent_student_directly(preverified=False)
        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium)
        page = page.go_to_independent_student_login_page()
        page = page.independent_student_login_failure(username, password)

        assert page.has_login_failed("independent_student_login_form", INVALID_LOGIN_MESSAGE)

        verify_email(page)

        assert is_email_verified_message_showing(self.selenium)

        page = page.independent_student_login(username, password)

        assert self.is_dashboard(page)

    def test_reset_password(self):
        page = self.go_to_homepage()

        page, name, username, _, _ = create_independent_student(page)
        page = self.get_to_forgotten_password_page()

        page.reset_email_submit(username)

        self.wait_for_email()

        page = email_utils.follow_reset_email_link(self.selenium, mail.outbox[0])

        new_password = "AnotherPassword12"

        page.student_reset_password(new_password)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_independent_student_login_page()
            .independent_student_login(username, new_password)
        )

        assert self.is_dashboard(page)

        page = page.go_to_account_page()
        assert page.check_account_details({"name": name})

    def test_reset_password_fail(self):
        page = self.get_to_forgotten_password_page()
        fake_email = "fake_email@fakeemail.com"
        page.reset_email_submit(fake_email)

        time.sleep(5)

        assert len(mail.outbox) == 0

    def test_update_name_success(self):
        homepage = self.go_to_homepage()

        play_page, name, student_username, _, password = create_independent_student(homepage)

        page = play_page.independent_student_login(student_username, password).go_to_account_page()

        assert page.check_account_details({"name": name})

        page = page.update_name_success("New name", password)

        assert self.is_dashboard(page)
        assert is_student_details_updated_message_showing(self.selenium)

    def test_update_name_failure(self):
        homepage = self.go_to_homepage()

        play_page, _, student_username, _, password = create_independent_student(homepage)

        page = (
            play_page.independent_student_login(student_username, password)
            .go_to_account_page()
            .update_name_failure("Name!", password)
        )

        assert self.is_account_page(page)
        assert page.was_form_invalid(
            "student_account_form", "Names may only contain letters, numbers, dashes, underscores, and spaces."
        )

    def test_change_email(self):
        homepage = self.go_to_homepage()

        _, _, _, student_email, password = create_independent_student(homepage)
        play_page, _, _, other_email, _ = create_independent_student(homepage)

        page = play_page.independent_student_login(student_email, password).go_to_account_page()

        # Try changing email to an existing email, should fail
        page = page.change_email(other_email, password)
        assert self.is_email_verification_page(page)
        assert is_student_details_updated_message_showing(self.selenium)
        assert is_email_updated_message_showing(self.selenium)

        subject = str(mail.outbox[0].subject)
        assert subject == "Duplicate account"
        mail.outbox = []

        # Try changing email to an existing teacher's email
        teacher_email, _ = signup_teacher_directly()

        page = (
            homepage.go_to_independent_student_login_page()
            .independent_student_login(student_email, password)
            .go_to_account_page()
        )

        page = page.change_email(teacher_email, password)
        assert self.is_email_verification_page(page)
        assert is_student_details_updated_message_showing(self.selenium)
        assert is_email_updated_message_showing(self.selenium)

        subject = str(mail.outbox[0].subject)
        assert subject == "Duplicate account"
        mail.outbox = []

        page = (
            self.go_to_homepage()
            .go_to_independent_student_login_page()
            .independent_student_login(student_email, password)
            .go_to_account_page()
        )

        # Try changing email to a new one, should succeed
        new_email = "another-email@codeforlife.com"
        page = page.change_email(new_email, password)

        assert self.is_email_verification_page(page)
        assert is_student_details_updated_message_showing(self.selenium)
        assert is_email_updated_message_showing(self.selenium)

        # Check user can still log in with old account before verifying new email
        self.selenium.get(self.live_server_url)
        page = (
            self.go_to_homepage()
            .go_to_independent_student_login_page()
            .independent_student_login(student_email, password)
        )
        assert self.is_dashboard(page)

        page = page.logout()

        page = email_utils.follow_change_email_link_to_independent_dashboard(page, mail.outbox[0])
        mail.outbox = []

        page = page.independent_student_login(new_email, password)

        assert self.is_dashboard(page)

    def test_change_password(self):
        homepage = self.go_to_homepage()

        play_page, _, student_username, _, password = create_independent_student(homepage)

        page = play_page.independent_student_login(student_username, password).go_to_account_page()

        new_password = "AnotherPassword12"
        page = page.update_password_success(new_password, password, is_independent=True)

        assert self.is_login_page(page)
        assert is_student_details_updated_message_showing(self.selenium)
        assert is_password_updated_message_showing(self.selenium)

        page = page.independent_student_login(student_username, new_password)

        assert self.is_dashboard(page)

    def test_join_class_nonexistent_class(self):
        homepage = self.go_to_homepage()

        play_page, _, student_username, _, password = create_independent_student(homepage)

        page = (
            play_page.independent_student_login(student_username, password)
            .go_to_join_a_school_or_club_page()
            .join_a_school_or_club_failure("AA123")
        )

        assert self.is_join_class_page(page)
        assert page.has_join_request_failed("Cannot find the school or club and/or class")

    def test_join_class_not_accepting_requests(self):
        teacher_email, _ = signup_teacher_directly()
        create_organisation_directly(teacher_email)
        _, _, access_code = create_class_directly(teacher_email)
        create_school_student_directly(access_code)

        homepage = self.go_to_homepage()

        play_page, _, student_username, _, password = create_independent_student(homepage)

        page = (
            play_page.independent_student_login(student_username, password)
            .go_to_join_a_school_or_club_page()
            .join_a_school_or_club_failure(access_code)
        )

        assert self.is_join_class_page(page)
        assert page.has_join_request_failed("Cannot find the school or club and/or class")

    def test_join_class_revoked(self):
        teacher_email, _ = signup_teacher_directly()
        create_organisation_directly(teacher_email)
        klass, _, access_code = create_class_directly(teacher_email)
        create_school_student_directly(access_code)
        klass.always_accept_requests = True
        klass.save()

        homepage = self.go_to_homepage()

        play_page, _, student_username, _, password = create_independent_student(homepage)

        page = (
            play_page.independent_student_login(student_username, password)
            .go_to_join_a_school_or_club_page()
            .join_a_school_or_club(access_code)
        )

        assert is_indep_student_join_request_received_message_showing(self.selenium)

        page.revoke_join_request()

        assert is_indep_student_join_request_revoked_message_showing(self.selenium)

    def test_join_class_accepted(self):
        teacher_email, teacher_password = signup_teacher_directly()
        create_organisation_directly(teacher_email)
        klass, _, access_code = create_class_directly(teacher_email)
        create_school_student_directly(access_code)
        klass.always_accept_requests = True
        klass.save()

        homepage = self.go_to_homepage()
        page = homepage.go_to_independent_student_login_page()

        username, password, student = create_independent_student_directly()

        page = (
            page.independent_student_login(username, password)
            .go_to_join_a_school_or_club_page()
            .join_a_school_or_club(access_code)
        )

        page.logout()

        page = self.go_to_homepage()

        page = (
            page.go_to_teacher_login_page()
            .login(teacher_email, teacher_password)
            .open_classes_tab()
            .accept_independent_join_request()
            .save(username)
            .return_to_class()
        )

        assert page.student_exists(username)

        # check whether a record is created correctly
        logs = JoinReleaseStudent.objects.filter(student=student)
        assert len(logs) == 1
        assert logs[0].action_type == JoinReleaseStudent.JOIN

    def test_join_class_denied(self):
        teacher_email, teacher_password = signup_teacher_directly()
        create_organisation_directly(teacher_email)
        klass, _, access_code = create_class_directly(teacher_email)
        create_school_student_directly(access_code)
        klass.always_accept_requests = True
        klass.save()

        homepage = self.go_to_homepage()

        play_page, _, student_username, _, password = create_independent_student(homepage)

        page = (
            play_page.independent_student_login(student_username, password)
            .go_to_join_a_school_or_club_page()
            .join_a_school_or_club(access_code)
        )

        page.logout()

        page = self.go_to_homepage()

        dashboard_page = (
            page.go_to_teacher_login_page()
            .login(teacher_email, teacher_password)
            .open_classes_tab()
            .deny_independent_join_request()
        )

        assert dashboard_page.has_no_independent_join_requests()

    def test_join_class_denied_and_accepted_by_admin(self):
        # Create 2 teachers in the same school, one admin, one standard
        admin_email, admin_password1 = signup_teacher_directly()
        standard_email, _ = signup_teacher_directly()
        name, postcode = create_organisation_directly(admin_email)
        join_teacher_to_organisation(standard_email, name, postcode, is_admin=False)

        # Create class for standard teacher which always accepts external requests
        klass, class_name, access_code = create_class_directly(standard_email)
        klass.always_accept_requests = True
        klass.save()

        # Create two independent students
        username1, password1, student1 = create_independent_student_directly()
        username2, password2, student2 = create_independent_student_directly()

        # Login as both students and request to join the same class
        homepage = self.go_to_homepage()
        page = homepage.go_to_independent_student_login_page()

        page = (
            page.independent_student_login(username1, password1)
            .go_to_join_a_school_or_club_page()
            .join_a_school_or_club(access_code)
        )

        page.logout()

        page = self.go_to_homepage()
        page = page.go_to_independent_student_login_page()

        page = (
            page.independent_student_login(username2, password2)
            .go_to_join_a_school_or_club_page()
            .join_a_school_or_club(access_code)
        )

        page.logout()

        # Login as school admin, accept the first request
        page = self.go_to_homepage()
        page = (
            page.go_to_teacher_login_page()
            .login(admin_email, admin_password1)
            .open_classes_tab()
            .accept_independent_join_request()
            .save(username1)
            .return_to_class()
        )

        assert page.student_exists(username1)

        # check whether a record is created correctly
        logs = JoinReleaseStudent.objects.filter(student=student1)
        assert len(logs) == 1
        assert logs[0].action_type == JoinReleaseStudent.JOIN

        # Deny the second request
        page = (
            page.go_to_dashboard()
            .open_classes_tab()
            .deny_independent_join_request()
        )

        assert page.has_no_independent_join_requests()

        # check a record hasn't been created and the student no longer has a join request
        logs = JoinReleaseStudent.objects.filter(student=student2)
        assert len(logs) == 0
        assert student2.pending_class_request is None
        assert student2.is_independent()

    def test_cannot_see_aimmo(self):
        page = self.go_to_homepage()
        page, _, username, _, password = create_independent_student(page)
        page = page.independent_student_login(username, password)

        assert page.element_does_not_exist_by_link_text("Kurono")

    def get_to_forgotten_password_page(self):
        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium).go_to_independent_student_login_page().go_to_indep_forgotten_password_page()
        return page

    def wait_for_email(self):
        WebDriverWait(self.selenium, 2).until(lambda driver: len(mail.outbox) == 1)

    def is_dashboard(self, page):
        return page.__class__.__name__ == "PlayDashboardPage"

    def is_account_page(self, page):
        return page.__class__.__name__ == "PlayAccountPage"

    def is_join_class_page(self, page):
        return page.__class__.__name__ == "JoinSchoolOrClubPage"

    def is_login_page(self, page):
        return page.__class__.__name__ == "IndependentStudentLoginPage"

    def is_email_verification_page(self, page):
        return page.__class__.__name__ == "EmailVerificationNeededPage"



from __future__ import absolute_import

from unittest.mock import patch
from common.tests.utils.organisation import create_organisation_directly, join_teacher_to_organisation
from common.tests.utils.teacher import signup_teacher_directly
from common.tests.utils.classes import create_class_directly
from common.tests.utils.student import create_school_student_directly

import pytest
from common.models import Class, School, Student, Teacher
from common.tests.utils.user import create_user_directly, get_superuser
from django.contrib.auth.models import User
from django.urls import reverse
from hamcrest import *
from hamcrest.core.base_matcher import BaseMatcher
from rest_framework import status
from rest_framework.test import APIClient, APITestCase


class APITests(APITestCase):
    def test_valid_date_registered(self):
        url = reverse("registered-users", kwargs={"year": "2016", "month": "04", "day": "01"})
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(isinstance(response.data, int))

    def test_invalid_date_registered(self):
        url = reverse("registered-users", kwargs={"year": "2016", "month": "05", "day": "35"})
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_404_NOT_FOUND))

    def test_valid_date_lastconnectedsince(self):
        url = reverse("last-connected-since", kwargs={"year": "2016", "month": "04", "day": "01"})
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(isinstance(response.data, int))

    def test_invalid_date_lastconnectedsince(self):
        url = reverse("last-connected-since", kwargs={"year": "2016", "month": "05", "day": "35"})
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_404_NOT_FOUND))

    def test_valid_country_userspercountry(self):
        url = reverse("number_users_per_country", kwargs={"country": "GB"})
        superuser = get_superuser()
        self.client.force_authenticate(user=superuser)
        response = self.client.get(url)
        assert_that(response, has_status_code(status.HTTP_200_OK))
        assert_that(isinstance(response.data, int))

    def test_get_inactive_users_if_admin(self):
        client = APIClient()
        superuser = get_superuser()
        create_user_directly(active=False)
        create_user_directly(active=True)
        client.force_authenticate(user=superuser)
        url = reverse("inactive_users")
        response = client.get(url)
        assert len(response.data) == 1

    @patch("portal.views.api.IS_CLOUD_SCHEDULER_FUNCTION", return_value=True)
    def test_get_inactive_users_if_appengine(self, mock_is_cloud_scheduler_function):
        client = APIClient()
        create_user_directly(active=False)
        create_user_directly(active=True)
        url = reverse("inactive_users")
        response = client.get(url)
        assert mock_is_cloud_scheduler_function.called
        assert response.status_code == status.HTTP_200_OK
        assert len(response.data) == 1

    def test_get_inactive_users_if_unauthorised(self):
        client = APIClient()
        create_user_directly(active=False)
        create_user_directly(active=True)
        url = reverse("inactive_users")
        response = client.get(url)
        assert response.status_code == status.HTTP_403_FORBIDDEN

    @patch("portal.views.api.IS_CLOUD_SCHEDULER_FUNCTION", return_value=True)
    def test_delete_inactive_users_if_appengine(self, mock_is_cloud_scheduler_function):
        client = APIClient()
        create_user_directly(active=False)
        create_user_directly(active=False)
        url = reverse("inactive_users")
        response = client.get(url)
        users = response.data
        assert len(users) == 2
        response = client.delete(url)
        assert mock_is_cloud_scheduler_function.called
        assert response.status_code == status.HTTP_204_NO_CONTENT
        for user in users:
            with pytest.raises(User.DoesNotExist):
                User.objects.get(username=user["username"])
        deleted_users = list(User.objects.filter(is_active=False))
        assert len(deleted_users) == 2
        for user in deleted_users:
            assert user.first_name == "Deleted"
            assert user.last_name == "User"
            assert user.email == ""
            assert not user.is_active
            assert not client.login(username=user.username, password="password")
        response = client.get(url)
        assert len(response.data) == 0

    @patch("portal.views.api.IS_CLOUD_SCHEDULER_FUNCTION", return_value=True)
    def test_orphan_schools_and_classes_are_anonymised(self, mock_is_cloud_scheduler_function):
        client = APIClient()
        # Create a school with an active teacher
        school1_teacher1_email, _ = signup_teacher_directly()
        school1_name, school1_postcode = create_organisation_directly(school1_teacher1_email)
        klass11, _, access_code11 = create_class_directly(school1_teacher1_email)
        _, _, student11 = create_school_student_directly(access_code11)

        # Create a school with one active non-admin teacher and one inactive admin teacher
        school2_teacher1_email, _ = signup_teacher_directly()
        school2_teacher2_email, _ = signup_teacher_directly()
        school2_name, school2_postcode = create_organisation_directly(school2_teacher1_email)
        join_teacher_to_organisation(school2_teacher2_email, school2_name, school2_postcode, is_admin=True)
        klass21, _, access_code21 = create_class_directly(school2_teacher1_email)
        _, _, student21 = create_school_student_directly(access_code21)
        klass22, _, access_code22 = create_class_directly(school2_teacher2_email)
        _, _, student22 = create_school_student_directly(access_code22)
        school2_teacher1 = Teacher.objects.get(new_user__email=school2_teacher1_email)
        school2_teacher1.is_admin = False
        school2_teacher1.save()
        school2_teacher2 = Teacher.objects.get(new_user__email=school2_teacher2_email)
        school2_teacher2.new_user.is_active = False
        school2_teacher2.new_user.save()

        # Create a school with 2 inactive teachers
        school3_teacher1_email, _ = signup_teacher_directly()
        school3_teacher2_email, _ = signup_teacher_directly()
        school3_name, school3_postcode = create_organisation_directly(school3_teacher1_email)
        join_teacher_to_organisation(school3_teacher2_email, school3_name, school3_postcode)
        klass31, _, access_code31 = create_class_directly(school3_teacher1_email)
        _, _, student31 = create_school_student_directly(access_code31)
        klass32, _, access_code32 = create_class_directly(school3_teacher2_email)
        _, _, student32 = create_school_student_directly(access_code32)
        school3_teacher1 = Teacher.objects.get(new_user__email=school3_teacher1_email)
        school3_teacher1.new_user.is_active = False
        school3_teacher1.new_user.save()
        school3_teacher2 = Teacher.objects.get(new_user__email=school3_teacher2_email)
        school3_teacher2.new_user.is_active = False
        school3_teacher2.new_user.save()

        # Create a school with no active teachers
        school4_teacher1_email, _ = signup_teacher_directly()
        school4_name, school4_postcode = create_organisation_directly(school4_teacher1_email)
        school4_teacher1 = Teacher.objects.get(new_user__email=school4_teacher1_email)
        school4_teacher1.new_user.is_active = False
        school4_teacher1.new_user.save()

        # Create a school with no teachers
        school5_teacher1_email, _ = signup_teacher_directly()
        school5_name, school5_postcode = create_organisation_directly(school5_teacher1_email)
        school5_teacher1 = Teacher.objects.get(new_user__email=school5_teacher1_email)
        school5_teacher1.delete()

        # Call the API
        url = reverse("anonymise_orphan_schools", kwargs={"start_id": 1})
        response = client.get(url)
        assert response.status_code == status.HTTP_204_NO_CONTENT

        # Check the first school/class/student still exist
        assert School.objects.filter(name=school1_name, postcode=school1_postcode).exists()
        assert Class.objects.filter(pk=klass11.pk).exists()
        assert Student.objects.filter(pk=student11.pk).exists()

        # Check the second school exists and its first class/student, but the second ones are anonymised
        assert School.objects.filter(name=school2_name, postcode=school2_postcode).exists()
        assert Class.objects.filter(pk=klass21.pk).exists()
        assert not Class.objects.filter(pk=klass22.pk).exists()
        assert Student.objects.filter(pk=student21.pk).exists()
        assert not Student.objects.get(pk=student22.pk).new_user.is_active
        # Also check the first teacher is now an admin
        assert Teacher.objects.get(new_user__email=school2_teacher1_email).is_admin

        # Check the third school is anonymised together with its classes and students
        assert not School.objects.filter(name=school3_name, postcode=school3_postcode).exists()
        assert not Class.objects.filter(pk=klass31.pk).exists()
        assert not Class.objects.filter(pk=klass32.pk).exists()
        assert not Student.objects.get(pk=student31.pk).new_user.is_active
        assert not Student.objects.get(pk=student32.pk).new_user.is_active

        # Check that the fourth school is anonymised
        assert not School.objects.filter(name=school4_name, postcode=school4_postcode).exists()

        # Check that the fifth school is anonymised
        assert not School.objects.filter(name=school5_name, postcode=school5_postcode).exists()


def has_status_code(status_code):
    return HasStatusCode(status_code)


class HasStatusCode(BaseMatcher):
    def __init__(self, status_code):
        self.status_code = status_code

    def _matches(self, response):
        return response.status_code == self.status_code

    def describe_to(self, description):
        description.append_text("has status code ").append_text(self.status_code)

    def describe_mismatch(self, response, mismatch_description):
        mismatch_description.append_text("had status code ").append_text(response.status_code)



from aimmo.models import Game
from common.models import Class
from django.core.exceptions import ValidationError
from django.db.models.query import QuerySet
from django.forms import ModelChoiceField, ModelForm, Select


class AddGameForm(ModelForm):
    def __init__(self, classes: QuerySet, *args, **kwargs):
        super(AddGameForm, self).__init__(*args, **kwargs)
        self.fields["game_class"].queryset = classes

    game_class = ModelChoiceField(
        queryset=None, widget=Select, label="Class", required=True
    )

    class Meta:
        model = Game
        fields = [
            "game_class",
        ]

    def clean(self):
        super(AddGameForm, self).clean()
        game_class: Class = self.cleaned_data.get("game_class")

        if not game_class:
            raise ValidationError("An invalid class was entered")

        if Game.objects.filter(game_class=game_class, is_archived=False).exists():
            raise ValidationError("An active game already exists for this class")

        return self.cleaned_data



from common.models import UserSession, Student, Class
from django.contrib import messages
from django.contrib.auth import authenticate, login
from django.contrib.auth.views import LoginView, FormView
from django.http import HttpResponseRedirect
from django.urls import reverse_lazy
from django.utils.html import escape

from portal.forms.play import StudentLoginForm, StudentClassCodeForm

import logging

LOGGER = logging.getLogger(__name__)


class StudentClassCodeView(FormView):
    template_name = "portal/login/student_class_code.html"
    form_class = StudentClassCodeForm
    extra_context = {"usertype": "SCHOOL_STUDENT"}

    def form_valid(self, form):
        self.form = form
        return HttpResponseRedirect(self.get_success_url())

    def get_success_url(self):
        class_code = self.form.cleaned_data["access_code"].upper()
        return reverse_lazy(
            "student_login",
            kwargs={"access_code": class_code, "login_type": "classform"},
        )


class StudentLoginView(LoginView):
    """View for login with class link"""

    template_name = "portal/login/student.html"
    form_class = StudentLoginForm
    success_url = reverse_lazy("student_details")
    extra_context = {"usertype": "SCHOOL_STUDENT"}

    def get_form_kwargs(self):
        kwargs = super(StudentLoginView, self).get_form_kwargs()
        kwargs["access_code"] = self.kwargs["access_code"].upper()
        return kwargs

    def _add_logged_in_as_message(self, request):
        class_name = self.kwargs["access_code"].upper()
        messages.info(
            request,
            f"<strong>You are logged in to class: " f"{escape(class_name)}</strong>",
            extra_tags="safe message--student",
        )

    def get_success_url(self):
        redirect_url = self.get_redirect_url()

        if redirect_url:
            return redirect_url

        self._add_logged_in_as_message(self.request)

        return self.success_url

    def _add_login_data(self, form, login_type):
        # class and student have been validated by this point
        class_code = self.kwargs["access_code"]
        classes = Class.objects.filter(access_code__iexact=class_code)
        klass = classes[0]

        name = form.cleaned_data.get("username")
        students = Student.objects.filter(
            new_user__first_name__iexact=name, class_field=klass
        )
        try:
            student = students[0]
        except IndexError:
            msg = f"Student {name} in class {class_code} is not found!"
            LOGGER.error(msg)
            raise Exception(msg)

        # Log the login time, class, and login type
        session = UserSession(
            user=student.new_user, class_field=klass, login_type=login_type
        )
        session.save()

    def form_valid(self, form):
        """Security check complete. Log the user in."""

        login_type = self.kwargs.get(
            "login_type", "classlink"
        )  # default to "classlink" if not specified

        self._add_login_data(form, login_type)
        return super(StudentLoginView, self).form_valid(form)


def student_direct_login(request, user_id, login_id):
    """Direct login for student with unique url without username and password"""
    user = authenticate(request, user_id=user_id, login_id=login_id)

    if user:
        # Log the login time and class
        student = Student.objects.get(new_user=user)
        session = UserSession(
            user=user, class_field=student.class_field, login_type="direct"
        )
        session.save()

        login(request, user)
        return HttpResponseRedirect(reverse_lazy("student_details"))
    return HttpResponseRedirect(reverse_lazy("home"))



from django.shortcuts import render
from django.core.exceptions import PermissionDenied


def reactTestSpace(request):
    if request.user.is_superuser:
        return render(request, "index.html")
    raise PermissionDenied()



from .base_page import BasePage
from .play import dashboard_page


class StudentLoginPage(BasePage):
    def __init__(self, browser):
        super(StudentLoginPage, self).__init__(browser)

        assert self.on_correct_page("student_login_page")

    def student_login(self, name, password):
        self._student_login(name, password)

        return dashboard_page.PlayDashboardPage(self.browser)

    def student_login_failure(self, name, password):
        self._student_login(name, password)
        return self

    def _student_login(self, name, password):
        self.browser.find_element_by_id("id_username").send_keys(name)
        self.browser.find_element_by_id("id_password").send_keys(password)
        self.browser.find_element_by_name("school_login").click()

    def has_login_failed(self, form_id, error):
        errors = (
            self.browser.find_element_by_id(form_id)
            .find_element_by_class_name("errorlist")
            .text
        )
        return error in errors



from django.middleware.security import SecurityMiddleware


class CustomSecurityMiddleware(SecurityMiddleware):
    """
    Extends Django's Security Middleware.
    See https://docs.djangoproject.com/en/2.2/_modules/django/middleware/security/ for
    the source code, as well as https://docs.djangoproject.com/en/2.2/ref/middleware/#module-django.middleware.security
    for docs on security middleware.
    """

    def process_response(self, request, response):
        """
        Extends the original security middleware to ensure the X-XSS-Protection header
        is set to 0.
        """
        super().process_response(request, response)

        if self.xss_filter:
            response["X-XSS-Protection"] = "0"

        return response



from django import forms


class NewsletterForm(forms.Form):
    email = forms.EmailField(
        label="Sign up to receive updates about Code for Life games and teaching resources.",
        label_suffix="",
        widget=forms.EmailInput(attrs={"placeholder": "Your email address", "id": "newsletter_email_field"}),
        help_text="Enter email address above",
    )

    age_verification = forms.BooleanField(widget=forms.CheckboxInput(), initial=False, required=True)


class ConsentForm(forms.Form):
    email = forms.EmailField(
        label="Email", label_suffix="", widget=forms.EmailInput(attrs={"placeholder": "your.name@yourdomain.com"})
    )

    consent_ticked = forms.BooleanField(widget=forms.CheckboxInput(), initial=False, required=True)



from django.urls import reverse

from portal.tests.base_test import BaseTest
from portal.tests.pageObjects.portal.teacher_login_page import TeacherLoginPage
from portal.views import admin


class TestAdmin(BaseTest):
    @classmethod
    def setUpClass(cls):
        super(TestAdmin, cls).setUpClass()
        admin.block_limit = 100

    # NB: Users are not expected to navigate to admin login page directly
    def navigate_to_admin_login(self):
        url = self.live_server_url + reverse("teacher_login")
        self.selenium.get(url)
        return TeacherLoginPage(self.selenium)

    # Checks all admin pages goes to admin_login when user is not logged in
    def test_navigate_to_admin_login(self):
        self.navigate_to_admin_login()



from __future__ import absolute_import

from . import class_page, dashboard_page
from .teach_base_page import TeachBasePage


class OnboardingStudentListPage(TeachBasePage):
    def __init__(self, browser):
        super(OnboardingStudentListPage, self).__init__(browser)

        assert self.on_correct_page("onboarding_student_list_page")

    def student_exists(self, name):
        return name in self.browser.find_element_by_id("student_table").text

    def is_student_password(self, password):
        return (
            password
            in self.browser.find_element_by_xpath(
                "//table[@id='student_table']/tbody/tr[4]/td[2]"
            ).text
        )

    def go_back_to_class(self):
        self.browser.find_element_by_id("back_to_class_button").click()
        return class_page.TeachClassPage(self.browser)

    def get_first_login_url(self):
        return self.browser.find_element_by_xpath(
            "//table[@id='student_table']/tbody/tr[4]/td[4]/div/div[1]"
        ).text

    def complete_setup(self):
        self.browser.find_element_by_id("complete_setup_button").click()
        return dashboard_page.TeachDashboardPage(self.browser)



from common.permissions import logged_in_as_teacher
from django.contrib.auth.decorators import login_required, user_passes_test
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse_lazy

from portal.strings.teacher_resources import (
    RAPID_ROUTER_RESOURCES_BANNER,
    KURONO_RESOURCES_BANNER,
)


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_rapid_router_resources(request):
    return render(
        request,
        "portal/teach/teacher_resources.html",
        {"BANNER": RAPID_ROUTER_RESOURCES_BANNER, "rapid_router_resources": True},
    )


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_kurono_resources(request):
    return render(
        request,
        "portal/teach/teacher_resources.html",
        {"BANNER": KURONO_RESOURCES_BANNER},
    )


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def materials(request):
    return HttpResponseRedirect(reverse_lazy("teaching_resources"))


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def kurono_teaching_packs(request):
    return HttpResponseRedirect(reverse_lazy("kurono_teaching_resources"))



AIMMO_DASHBOARD_BANNER = {
    "title": "Kurono dashboard",
    "subtitle": "Let the adventures begin! A collection of challenges await your "
    "exploration. See how far you can get.",
    "image_class": "banner--picture--aimmo",
}



from common.models import Class, EmailVerification, School, SchoolTeacherInvitation, Student, Teacher, UserProfile
from django.contrib import admin
from django.contrib.auth.admin import UserAdmin
from django.contrib.auth.models import User

from portal.forms.admin import AdminChangeUserPasswordForm, AdminUserCreationForm
from portal.views.api import anonymise


class ClassAdmin(admin.ModelAdmin):
    search_fields = ["name", "teacher__new_user__first_name", "teacher__new_user__last_name", "teacher__school__name"]
    list_display = ["__str__", "teacher", "teacher_school"]
    raw_id_fields = ["teacher"]

    def teacher_school(self, obj):
        return obj.teacher.school


class SchoolAdmin(admin.ModelAdmin):
    search_fields = ["name", "country", "postcode", "town"]
    list_filter = ["postcode", "country"]


class StudentAdmin(admin.ModelAdmin):
    search_fields = [
        "new_user__first_name",
        "new_user__last_name",
        "class_field__name",
        "class_field__teacher__new_user__first_name",
        "class_field__teacher__new_user__last_name",
        "class_field__teacher__school__name",
    ]
    list_display = ["__str__", "class_field", "class_field_teacher", "class_field_school"]
    readonly_fields = ["user", "new_user"]
    raw_id_fields = ["class_field", "pending_class_request"]

    def class_field_teacher(self, obj):
        if obj.class_field:
            return obj.class_field.teacher
        else:
            return "Independent"

    def class_field_school(self, obj):
        if obj.class_field:
            return obj.class_field.teacher.school
        else:
            return "Independent"


class TeacherAdmin(admin.ModelAdmin):
    search_fields = ["new_user__first_name", "new_user__last_name", "school__name"]
    list_display = ["__str__", "school"]
    readonly_fields = ["user", "new_user"]
    raw_id_fields = ["school"]


class UserProfileAdmin(admin.ModelAdmin):
    search_fields = ["user__first_name", "user__last_name", "user__username", "user__date_joined"]
    list_filter = ["user__date_joined"]
    list_display = ["user", "__str__", "joined_recently"]
    readonly_fields = ["user"]


class EmailVerificationAdmin(admin.ModelAdmin):
    search_fields = ["user__first_name", "user__last_name", "user__username", "user__date_joined"]
    readonly_fields = ["user", "token"]


class SchoolTeacherInvitationAdmin(admin.ModelAdmin):
    search_fields = [
        "from_teacher__new_user__first_name",
        "from_teacher__new_user__last_name",
        "from_teacher__new_user__email",
        "school__name",
        "invited_teacher_first_name",
        "invited_teacher_last_name",
        "invited_teacher_email",
        "expiry",
        "creation_time",
    ]
    readonly_fields = ["token", "school", "from_teacher"]


def anonymise_user(user_admin, request, queryset):
    for user in queryset:
        anonymise(user)


anonymise_user.short_description = "Anonymise selected users"


UserAdmin.list_display += ("date_joined",)
UserAdmin.list_filter += ("date_joined",)
UserAdmin.add_form = AdminUserCreationForm
UserAdmin.change_password_form = AdminChangeUserPasswordForm
UserAdmin.actions.append(anonymise_user)


admin.site.register(Class, ClassAdmin)
admin.site.register(Student, StudentAdmin)
admin.site.register(Teacher, TeacherAdmin)
admin.site.register(School, SchoolAdmin)
admin.site.unregister(User)
admin.site.register(User, UserAdmin)
admin.site.register(UserProfile, UserProfileAdmin)
admin.site.register(EmailVerification, EmailVerificationAdmin)
admin.site.register(SchoolTeacherInvitation, SchoolTeacherInvitationAdmin)



import re
from datetime import timedelta, date

from captcha.fields import ReCaptchaField
from captcha.widgets import ReCaptchaV2Invisible
from common.helpers.emails import send_verification_email
from common.models import Class, Student, stripStudentName
from common.permissions import logged_in_as_independent_student
from django import forms
from django.contrib.auth import authenticate
from django.contrib.auth.forms import AuthenticationForm
from django.utils import timezone

from portal.forms.error_messages import INVALID_LOGIN_MESSAGE
from portal.helpers.password import PasswordStrength, form_clean_password
from portal.helpers.regexes import ACCESS_CODE_PATTERN
from portal.templatetags.app_tags import is_verified


class StudentClassCodeForm(forms.Form):
    access_code = forms.CharField(
        widget=forms.TextInput(attrs={"autocomplete": "off", "placeholder": "Class code"}),
        help_text="Enter your class code",
    )

    def clean(self):
        access_code = self.cleaned_data.get("access_code", None)

        if access_code:
            if re.fullmatch(ACCESS_CODE_PATTERN, access_code.upper()) is None:
                raise forms.ValidationError("Uh oh! You didn't input a valid class code.")

        return self.cleaned_data


class StudentLoginForm(AuthenticationForm):
    username = forms.CharField(
        widget=forms.TextInput(attrs={"autocomplete": "off", "placeholder": "Username"}),
        help_text="Enter your username",
    )
    password = forms.CharField(
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Password"}),
        help_text="Enter your password",
    )

    def __init__(self, *args, **kwargs):
        self.access_code = kwargs.pop("access_code", None)
        super(StudentLoginForm, self).__init__(*args, **kwargs)

    def clean(self):
        name = self.cleaned_data.get("username", None)
        password = self.cleaned_data.get("password", None)

        if name and self.access_code and password:

            student, user = self.check_for_errors(name, self.access_code, password)

            self.student = student
            self.user_cache = user
        return self.cleaned_data

    def check_for_errors(self, name, access_code, password):
        classes = Class.objects.filter(access_code__iexact=access_code)
        if len(classes) != 1:
            raise forms.ValidationError("Invalid name, class access code or password")
        klass = classes[0]

        name = stripStudentName(name)

        students = Student.objects.filter(new_user__first_name__iexact=name, class_field=klass)
        if len(students) != 1:
            raise forms.ValidationError("Invalid name, class access code or password")

        student = students[0]
        user = authenticate(username=student.new_user.username, password=password.lower())

        # Try the case sensitive password too, for previous accounts that don't have the lowercase one stored
        if user is None:
            user = authenticate(username=student.new_user.username, password=password)

        if user is None:
            raise forms.ValidationError("Invalid name, class access code or password")
        if not user.is_active:
            raise forms.ValidationError("This user account has been deactivated")

        return student, user


class StudentEditAccountForm(forms.Form):
    password = forms.CharField(
        required=True, widget=forms.PasswordInput(attrs={"placeholder": "New password"}), help_text="Enter new password"
    )
    confirm_password = forms.CharField(
        required=True,
        widget=forms.PasswordInput(attrs={"placeholder": "Confirm new password"}),
        help_text="Confirm new password",
    )
    current_password = forms.CharField(
        widget=forms.PasswordInput(attrs={"placeholder": "Current password"}), help_text="Enter your current password"
    )

    def __init__(self, user, *args, **kwargs):
        self.user = user
        super(StudentEditAccountForm, self).__init__(*args, **kwargs)

    def clean_password(self):
        return form_clean_password(self, "password", PasswordStrength.STUDENT)

    def clean(self):
        return clean_confirm_password(self, independent=False)


class IndependentStudentEditAccountForm(forms.Form):
    name = forms.CharField(
        max_length=100,
        required=False,
        widget=forms.TextInput(attrs={"placeholder": "Name"}),
        help_text="Enter your name",
    )
    email = forms.EmailField(
        required=False,
        widget=forms.EmailInput(attrs={"placeholder": "New email address (optional)"}),
        help_text="Enter new email address (optional)",
    )
    password = forms.CharField(
        required=False,
        widget=forms.PasswordInput(attrs={"placeholder": "New password (optional)"}),
        help_text="Enter new password (optional)",
    )
    confirm_password = forms.CharField(
        label="Confirm new password",
        required=False,
        widget=forms.PasswordInput(attrs={"placeholder": "Confirm new password"}),
        help_text="Confirm new password",
    )
    current_password = forms.CharField(
        label="Current password",
        widget=forms.PasswordInput(attrs={"placeholder": "Current password"}),
        help_text="Enter your current password",
    )

    def __init__(self, user, *args, **kwargs):
        self.user = user
        super(IndependentStudentEditAccountForm, self).__init__(*args, **kwargs)

    def clean_name(self):
        name = self.cleaned_data.get("name", None)
        if not self.user.new_student.class_field:
            if name == "":
                raise forms.ValidationError("This field is required")

            if re.match(re.compile("^[\w ]+$"), name) is None:
                raise forms.ValidationError("Names may only contain letters, numbers, dashes, underscores, and spaces.")

        return name

    def clean_password(self):
        return form_clean_password(self, "password", PasswordStrength.INDEPENDENT)

    def clean(self):
        return clean_confirm_password(self, independent=True)


def clean_confirm_password(self, independent=True):
    password = self.cleaned_data.get("password", None)
    confirm_password = self.cleaned_data.get("confirm_password", None)
    current_password = self.cleaned_data.get("current_password", None)

    # Password is lowercase for non-independent students
    if not independent:
        if password is not None:
            password = password.lower()
        if confirm_password is not None:
            confirm_password = confirm_password.lower()

    if are_password_and_confirm_password_different(password, confirm_password):
        raise forms.ValidationError("Your new passwords do not match")

    if current_password and not self.user.check_password(current_password):
        # If it's not an independent student, check their lowercase password as well
        if independent or not self.user.check_password(current_password.lower()):
            raise forms.ValidationError("Your current password was incorrect")

    return self.cleaned_data


def are_password_and_confirm_password_different(password, confirm_password):
    return password is not None and password != confirm_password


class IndependentStudentSignupForm(forms.Form):
    date_of_birth = forms.DateField(
        help_text="Please enter your date of birth (we do not store this information).",
        widget=forms.SelectDateWidget(
            years=range(date.today().year, date.today().year - 100, -1), empty_label=("Year", "Month", "Day")
        ),
        required=False,
    )

    name = forms.CharField(
        max_length=100,
        help_text="Enter full name",
        widget=forms.TextInput(attrs={"autocomplete": "off", "placeholder": "Full name"}),
    )

    email = forms.EmailField(
        help_text="Enter your email address",
        widget=forms.EmailInput(attrs={"autocomplete": "off", "placeholder": "Email address"}),
    )

    consent_ticked = forms.BooleanField(widget=forms.CheckboxInput(), initial=False, required=True)
    newsletter_ticked = forms.BooleanField(widget=forms.CheckboxInput(), initial=False, required=False)

    password = forms.CharField(
        help_text="Enter a password",
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Password"}),
    )

    confirm_password = forms.CharField(
        help_text="Repeat password",
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Repeat password"}),
    )

    captcha = ReCaptchaField(widget=ReCaptchaV2Invisible)

    def clean_name(self):
        name = self.cleaned_data.get("name", None)
        if re.match(re.compile("^[\w ]+$"), name) is None:
            raise forms.ValidationError("Names may only contain letters, numbers, dashes, underscores, and spaces.")

        return name

    def clean_password(self):
        return form_clean_password(self, "password", PasswordStrength.INDEPENDENT)

    def clean(self):
        password = self.cleaned_data.get("password", None)
        confirm_password = self.cleaned_data.get("confirm_password", None)

        if password and confirm_password and password != confirm_password:
            raise forms.ValidationError("Your passwords do not match")

        return self.cleaned_data


class IndependentStudentLoginForm(AuthenticationForm):
    username = forms.EmailField(
        widget=forms.EmailInput(attrs={"autocomplete": "off", "placeholder": "Email address"}),
        help_text="Enter your email address",
    )
    password = forms.CharField(
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Password"}),
        help_text="Enter your password",
    )

    def clean(self):
        super().clean()

    def confirm_login_allowed(self, user):
        if not logged_in_as_independent_student(user):
            self.show_invalid_login_message()

        if not is_verified(user):
            send_verification_email(self.request, user)
            self.show_invalid_login_message()

    def get_invalid_login_error(self):
        self.show_invalid_login_message()

    def show_invalid_login_message(self):
        raise forms.ValidationError(INVALID_LOGIN_MESSAGE)


class StudentJoinOrganisationForm(forms.Form):
    access_code = forms.CharField(label="Class Access Code", widget=forms.TextInput(attrs={"placeholder": "AB123"}))

    def clean(self):
        access_code = self.cleaned_data.get("access_code", None)

        if access_code:
            classes = Class.objects.filter(access_code=access_code)
            if len(classes) != 1:
                raise forms.ValidationError("Cannot find the school or club and/or class")
            self.klass = classes[0]
            if not self.klass.always_accept_requests:
                if self.klass.accept_requests_until is None:
                    raise forms.ValidationError("Cannot find the school or club and/or class")
                elif (self.klass.accept_requests_until - timezone.now()) < timedelta():
                    raise forms.ValidationError("Cannot find the school or club and/or class")
        return self.cleaned_data



from .base_test import BaseTest
from time import sleep
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from common.models import DailyActivity

from datetime import timedelta, datetime


class TestCodingClubs(BaseTest):
    def test_coding_club_increment(self):

        # first create dailyActivity one day before datetime.now()
        # to check if it can handle incrementing on different days
        # then check if increments are done on the same day
        old_date = datetime.now() - timedelta(days=1)
        old_daily_activity = DailyActivity(date=old_date)
        old_daily_activity.save()

        for i in range(4):
            # check both buttons
            self.go_to_homepage()
            button_id = "primary_pack" if i < 2 else "python_pack"
            find_out_more_button = WebDriverWait(self.selenium, 10).until(
                EC.element_to_be_clickable((By.ID, "find_out_more"))
            )
            find_out_more_button.click()

            daily_count_button = WebDriverWait(self.selenium, 10).until(
                EC.visibility_of_element_located((By.ID, button_id))
            )
            daily_count_button.click()
        # check the old_date is still the same
        old_daily_activity = DailyActivity.objects.get(date=old_date)
        assert old_daily_activity.primary_coding_club_downloads == 0
        assert old_daily_activity.python_coding_club_downloads == 0
        # check the current_date is incremented to 2
        current_daily_activity = DailyActivity.objects.get(date=datetime.now())
        assert current_daily_activity.primary_coding_club_downloads == 2
        assert current_daily_activity.python_coding_club_downloads == 2



from django.shortcuts import render


def privacy_policy(request):
    return render(request, "portal/privacy_policy.html", {"last_updated": "11th July 2022"})


def terms(request):
    return render(request, "portal/terms.html", {"last_updated": "11th July 2022"})



from __future__ import absolute_import

import re

import time
from aimmo.models import Game
from common.models import Class, Student, Teacher
from common.tests.utils import email as email_utils
from common.tests.utils.classes import create_class_directly
from common.tests.utils.organisation import create_organisation_directly, join_teacher_to_organisation
from common.tests.utils.student import create_independent_student_directly, create_school_student_directly
from common.tests.utils.teacher import (
    signup_duplicate_teacher_fail,
    signup_teacher,
    signup_teacher_directly,
    verify_email,
)
from django.core import mail
from django.test import Client, TestCase
from django.urls import reverse
from selenium.webdriver.common.by import By
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC

from portal.forms.error_messages import INVALID_LOGIN_MESSAGE
from portal.tests.test_invite_teacher import FADE_TIME
from .base_test import BaseTest
from .pageObjects.portal.home_page import HomePage
from .utils.messages import (
    is_message_showing,
    is_email_verified_message_showing,
    is_teacher_details_updated_message_showing,
    is_email_updated_message_showing,
    is_password_updated_message_showing,
)

from portal.tests.test_invite_teacher import WAIT_TIME


class TestTeacher(TestCase):
    def test_new_student_can_play_games(self):
        """
        Given a teacher has an kurono game,
        When they add a new student to their class,
        Then the new student should be able to play that class's games
        """
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        klass, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        c = Client()
        c.login(username=email, password=password)
        c.post(reverse("teacher_aimmo_dashboard"), {"game_class": klass.id})
        c.post(reverse("view_class", kwargs={"access_code": access_code}), {"names": "Florian"})

        game = Game.objects.get(id=1)
        new_student = Student.objects.last()
        assert game.can_user_play(new_student.new_user)

    def test_accepted_independent_student_can_play_games(self):
        """
        Given an independent student requests access to a class,
        When the teacher for that class accepts the request,
        Then the new student should have access to that class's games
        """
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        klass, _, access_code = create_class_directly(email)
        klass.always_accept_requests = True
        klass.save()
        create_school_student_directly(access_code)
        (indep_username, indep_password, indep_student) = create_independent_student_directly()

        c = Client()

        c.login(username=indep_username, password=indep_password)
        c.post(reverse("student_join_organisation"), {"access_code": access_code, "class_join_request": "Request"})
        c.logout()

        c.login(username=email, password=password)
        c.post(reverse("teacher_aimmo_dashboard"), {"game_class": klass.pk})
        c.post(reverse("teacher_accept_student_request", kwargs={"pk": indep_student.pk}), {"name": "Florian"})

        game: Game = Game.objects.get(id=1)
        new_student = Student.objects.last()
        assert game.can_user_play(new_student.new_user)

    def test_moved_class_has_correct_permissions_for_students_and_teachers(self):
        """
        Given two teachers each with a class and an aimmo game,
        When teacher 1 transfers their class to teacher 2,
        Then:
            - Students in each class still only have access to their class games
            - Teacher 2 has access to both games and teacher 1 has access to none
        """

        # Create teacher 1 -> class 1 -> student 1
        email1, password1 = signup_teacher_directly()
        school_name, postcode = create_organisation_directly(email1)
        klass1, _, access_code1 = create_class_directly(email1, "Class 1")
        create_school_student_directly(access_code1)

        # Create teacher 2 -> class 2 -> student 2
        email2, password2 = signup_teacher_directly()
        join_teacher_to_organisation(email2, school_name, postcode)
        klass2, _, access_code2 = create_class_directly(email2, "Class 2")
        create_school_student_directly(access_code2)

        teacher1: Teacher = Teacher.objects.get(new_user__email=email1)
        teacher2: Teacher = Teacher.objects.get(new_user__email=email2)

        c = Client()

        # Create game 1 under class 1
        c.login(username=email1, password=password1)
        c.post(reverse("teacher_aimmo_dashboard"), {"game_class": klass1.pk})
        c.logout()

        # Create game 2 under class 2
        c.login(username=email2, password=password2)
        c.post(reverse("teacher_aimmo_dashboard"), {"game_class": klass2.pk})
        c.logout()

        game1: Game = Game.objects.get(owner=teacher1.new_user)
        game2: Game = Game.objects.get(owner=teacher2.new_user)

        student1: Student = Student.objects.get(class_field=klass1)
        student2: Student = Student.objects.get(class_field=klass2)

        # Check student permissions for each game
        assert game1.can_user_play(student1.new_user)
        assert game2.can_user_play(student2.new_user)
        assert not game1.can_user_play(student2.new_user)
        assert not game2.can_user_play(student1.new_user)

        # Check teacher permissions for each game
        assert game1.can_user_play(teacher1.new_user)
        assert game2.can_user_play(teacher2.new_user)
        assert not game1.can_user_play(teacher2.new_user)
        assert not game2.can_user_play(teacher1.new_user)

        # Transfer class 1 over to teacher 2
        c.login(username=email1, password=password1)
        response = c.post(
            reverse("teacher_edit_class", kwargs={"access_code": access_code1}),
            {"new_teacher": teacher2.pk, "class_move_submit": ""},
        )
        assert response.status_code == 302
        c.logout()

        # Refresh model instances
        klass1: Class = Class.objects.get(pk=klass1.pk)
        klass2: Class = Class.objects.get(pk=klass2.pk)
        game1 = Game.objects.get(pk=game1.pk)
        game2 = Game.objects.get(pk=game2.pk)

        # Check teacher 2 is the teacher for class 1
        assert klass1.teacher == teacher2

        # Check that the students' permissions have not changed
        assert game1.can_user_play(student1.new_user)
        assert game2.can_user_play(student2.new_user)
        assert not game1.can_user_play(student2.new_user)
        assert not game2.can_user_play(student1.new_user)

        # Check that teacher 1 cannot access class 1's game 1 anymore
        assert not game1.can_user_play(teacher1.new_user)

        # Check that teacher 2 can access game 1
        assert game1.can_user_play(teacher2.new_user)

    def test_moved_student_has_access_to_only_new_teacher_games(self):
        """
        Given a student in a class,
        When a teacher transfers them to another class with a new teacher,
        Then the student should only have access to the new teacher's games
        """

        email1, password1 = signup_teacher_directly()
        school_name, postcode = create_organisation_directly(email1)
        klass1, _, access_code1 = create_class_directly(email1, "Class 1")
        create_school_student_directly(access_code1)

        email2, password2 = signup_teacher_directly()
        join_teacher_to_organisation(email2, school_name, postcode)
        klass2, _, access_code2 = create_class_directly(email2, "Class 2")
        create_school_student_directly(access_code2)

        teacher1 = Teacher.objects.get(new_user__email=email1)
        teacher2 = Teacher.objects.get(new_user__email=email2)

        c = Client()
        c.login(username=email2, password=password2)
        c.post(reverse("teacher_aimmo_dashboard"), {"game_class": klass2.pk})
        c.logout()

        c.login(username=email1, password=password1)
        c.post(reverse("teacher_aimmo_dashboard"), {"game_class": klass1.pk})

        game1 = Game.objects.get(owner=teacher1.new_user)
        game2 = Game.objects.get(owner=teacher2.new_user)

        student1 = Student.objects.get(class_field=klass1)
        student2 = Student.objects.get(class_field=klass2)

        assert game1.can_user_play(student1.new_user)
        assert game2.can_user_play(student2.new_user)

        c.post(
            reverse("teacher_move_students", kwargs={"access_code": access_code1}), {"transfer_students": student1.pk}
        )
        c.post(
            reverse("teacher_move_students_to_class", kwargs={"access_code": access_code1}),
            {
                "form-0-name": student1.user.user.first_name,
                "form-MAX_NUM_FORMS": 1000,
                "form-0-orig_name": student1.user.user.first_name,
                "form-TOTAL_FORMS": 1,
                "form-MIN_NUM_FORMS": 0,
                "submit_disambiguation": "",
                "form-INITIAL_FORMS": 1,
                "new_class": klass2.pk,
            },
        )
        c.logout()

        game1 = Game.objects.get(owner=teacher1.new_user)
        game2 = Game.objects.get(owner=teacher2.new_user)

        assert not game1.can_user_play(student1.new_user)
        assert game2.can_user_play(student1.new_user)

    def test_teacher_cannot_create_duplicate_game(self):
        """
        Given a teacher, a class and a worksheet,
        When the teacher creates a game for that class and worksheet, and then tries to
        create the exact same game again,
        Then the class should only have one game, and an error message should appear.
        """

        email, password = signup_teacher_directly()
        _, _ = create_organisation_directly(email)
        klass, _, _ = create_class_directly(email)

        c = Client()
        c.login(username=email, password=password)
        game1_response = c.post(reverse("teacher_aimmo_dashboard"), {"game_class": klass.pk})

        assert game1_response.status_code == 302
        assert Game.objects.filter(game_class=klass, is_archived=False).count() == 1
        assert klass.active_game != None
        messages = list(game1_response.wsgi_request._messages)
        assert len([m for m in messages if m.tags == "warning"]) == 0

        game2_response = c.post(reverse("teacher_aimmo_dashboard"), {"game_class": klass.pk})

        messages = list(game2_response.wsgi_request._messages)
        assert len([m for m in messages if m.tags == "warning"]) == 1
        assert messages[0].message == "An active game already exists for this class"

    def test_signup_short_password_fails(self):
        c = Client()

        response = c.post(
            reverse("register"),
            {
                "teacher_signup-teacher_first_name": "Test Name",
                "teacher_signup-teacher_last_name": "Test Last Name",
                "teacher_signup-teacher_email": "test@email.com",
                "teacher_signup-consent_ticked": "on",
                "teacher_signup-teacher_password": "test",
                "teacher_signup-teacher_confirm_password": "test",
                "g-recaptcha-response": "something",
            },
        )

        # Assert response isn't a redirect (submit failure)
        assert response.status_code == 200

    def test_signup_common_password_fails(self):
        c = Client()

        response = c.post(
            reverse("register"),
            {
                "teacher_signup-teacher_first_name": "Test Name",
                "teacher_signup-teacher_last_name": "Test Last Name",
                "teacher_signup-teacher_email": "test@email.com",
                "teacher_signup-consent_ticked": "on",
                "teacher_signup-teacher_password": "Password1",
                "teacher_signup-teacher_confirm_password": "Password1",
                "g-recaptcha-response": "something",
            },
        )

        # Assert response isn't a redirect (submit failure)
        assert response.status_code == 200

    def test_signup_passwords_do_not_match_fails(self):
        c = Client()

        response = c.post(
            reverse("register"),
            {
                "teacher_signup-teacher_first_name": "Test Name",
                "teacher_signup-teacher_last_name": "Test Last Name",
                "teacher_signup-teacher_email": "test@email.com",
                "teacher_signup-consent_ticked": "on",
                "teacher_signup-teacher_password": "StrongPassword1!",
                "teacher_signup-teacher_confirm_password": "StrongPassword2!",
                "g-recaptcha-response": "something",
            },
        )

        # Assert response isn't a redirect (submit failure)
        assert response.status_code == 200

    def test_signup_fails_without_consent(self):
        c = Client()

        response = c.post(
            reverse("register"),
            {
                "teacher_signup-teacher_first_name": "Test Name",
                "teacher_signup-teacher_last_name": "Test Last Name",
                "teacher_signup-teacher_email": "test@email.com",
                "teacher_signup-teacher_password": "StrongPassword1!",
                "teacher_signup-teacher_confirm_password": "StrongPassword1!",
                "g-recaptcha-response": "something",
            },
        )

        # Assert response isn't a redirect (submit failure)
        assert response.status_code == 200

    def test_signup_email_verification(self):
        c = Client()

        response = c.post(
            reverse("register"),
            {
                "teacher_signup-teacher_first_name": "Test Name",
                "teacher_signup-teacher_last_name": "Test Last Name",
                "teacher_signup-teacher_email": "test@email.com",
                "teacher_signup-consent_ticked": "on",
                "teacher_signup-teacher_password": "StrongPassword1!",
                "teacher_signup-teacher_confirm_password": "StrongPassword1!",
                "g-recaptcha-response": "something",
            },
        )

        assert response.status_code == 302
        assert len(mail.outbox) == 1

        # Try verification URL with a fake token
        bad_url = reverse("verify_email", kwargs={"token": "abcdef"})
        bad_verification_response = c.get(bad_url)

        # Assert response isn't a redirect (get failure)
        assert bad_verification_response.status_code == 200

        # Get verification link from email
        message = str(mail.outbox[0].body)
        verification_url = re.search("http.+/", message).group(0)

        # Verify the email properly
        verification_response = c.get(verification_url)

        # Assert response redirects and succeeds
        assert verification_response.status_code == 302

        # Try verifying the email a second time
        second_verification_response = c.get(verification_url)

        # Assert response isn't a redirect (get failure)
        assert second_verification_response.status_code == 200


# Class for Selenium tests. We plan to replace these and turn them into Cypress tests
class TestTeacherFrontend(BaseTest):
    def test_signup_without_newsletter(self):
        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium)
        page, _, _ = signup_teacher(page)
        assert is_email_verified_message_showing(self.selenium)

    def test_signup_with_newsletter(self):
        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium)
        page, _, _ = signup_teacher(page, newsletter=True)
        assert is_email_verified_message_showing(self.selenium)

    def test_signup_duplicate_failure(self):
        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium)
        page, email, _ = signup_teacher(page)
        assert is_email_verified_message_showing(self.selenium)

        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium)
        page, _, _ = signup_duplicate_teacher_fail(page, email)
        assert self.is_login_page(page)

        # Test sign up with an existing indy student's email
        indy_email, _, _ = create_independent_student_directly()
        page = self.go_to_homepage()
        page, _, _ = signup_duplicate_teacher_fail(page, indy_email)

    def test_login_failure(self):
        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium)
        page = page.go_to_teacher_login_page()
        page = page.login_failure("non-existent-email@codeforlife.com", "Incorrect password")
        assert page.has_login_failed("form-login-teacher", INVALID_LOGIN_MESSAGE)

    def test_login_success(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)
        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium)
        page = page.go_to_teacher_login_page()
        page = page.login(email, password)
        assert self.is_dashboard_page(page)

    def test_login_not_verified(self):
        email, password = signup_teacher_directly(preverified=False)
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)
        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium)
        page = page.go_to_teacher_login_page()
        page = page.login_failure(email, password)

        assert page.has_login_failed("form-login-teacher", INVALID_LOGIN_MESSAGE)

        verify_email(page)

        assert is_email_verified_message_showing(self.selenium)

        page = page.login(email, password)

        assert self.is_dashboard_page(page)

    def test_signup_login_success(self):
        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium)
        page, email, password = signup_teacher(page)
        page = page.login_no_school(email, password)
        assert self.is_onboarding_page(page)

    def test_view_resources(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)
        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium)
        page = page.go_to_teacher_login_page()
        page = page.login(email, password)

        assert self.is_dashboard_page(page)

        page = page.go_to_rapid_router_resources_page()

        assert self.is_resources_page(page)

        page = page.go_to_kurono_resources_page()

        assert self.is_resources_page(page)

    def test_edit_details(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium).go_to_teacher_login_page().login(email, password).open_account_tab()

        page = page.change_teacher_details(
            {"first_name": "Paulina", "last_name": "Koch", "current_password": "Password2!"}
        )
        assert self.is_dashboard_page(page)
        assert is_teacher_details_updated_message_showing(self.selenium)

        assert page.check_account_details({"first_name": "Paulina", "last_name": "Koch"})

    def test_edit_details_non_admin(self):
        email_1, _ = signup_teacher_directly()
        email_2, password_2 = signup_teacher_directly()
        name, postcode = create_organisation_directly(email_1)
        _, _, access_code_1 = create_class_directly(email_1)
        create_school_student_directly(access_code_1)
        join_teacher_to_organisation(email_2, name, postcode)
        _, _, access_code_2 = create_class_directly(email_2)
        create_school_student_directly(access_code_2)

        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium).go_to_teacher_login_page().login(email_2, password_2).open_account_tab()

        page = page.change_teacher_details(
            {"first_name": "Florian", "last_name": "Aucomte", "current_password": "Password2!"}
        )
        assert self.is_dashboard_page(page)
        assert is_teacher_details_updated_message_showing(self.selenium)

        assert page.check_account_details({"first_name": "Florian", "last_name": "Aucomte"})

    def test_change_email(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        other_email, _ = signup_teacher_directly()

        page = self.go_to_homepage()
        page = page.go_to_teacher_login_page().login(email, password).open_account_tab()

        # Try changing email to an existing email, should fail
        page = page.change_email("Test", "Teacher", other_email, password)
        assert self.is_email_verification_page(page)
        assert is_email_updated_message_showing(self.selenium)

        subject = str(mail.outbox[0].subject)
        assert subject == "Duplicate account"
        mail.outbox = []

        # Try changing email to an existing indy student's email, should fail
        indy_email, _, _ = create_independent_student_directly()
        page = self.go_to_homepage()
        page = page.go_to_teacher_login_page().login(email, password).open_account_tab()

        page = page.change_email("Test", "Teacher", indy_email, password)
        assert self.is_email_verification_page(page)
        assert is_email_updated_message_showing(self.selenium)

        subject = str(mail.outbox[0].subject)
        assert subject == "Duplicate account"
        mail.outbox = []

        page = self.go_to_homepage()
        page = page.go_to_teacher_login_page().login(email, password).open_account_tab()

        # Try changing email to a new one, should succeed
        new_email = "another-email@codeforlife.com"
        page = page.change_email("Test", "Teacher", new_email, password)
        assert self.is_email_verification_page(page)
        assert is_email_updated_message_showing(self.selenium)

        # Check user can still log in with old account before verifying new email
        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium).go_to_teacher_login_page().login(email, password)
        assert self.is_dashboard_page(page)

        page = page.logout()

        page = email_utils.follow_change_email_link_to_dashboard(page, mail.outbox[0])
        mail.outbox = []

        page = page.login(new_email, password).open_account_tab()

        assert page.check_account_details({"first_name": "Test", "last_name": "Teacher"})

    def test_change_password(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium).go_to_teacher_login_page().login(email, password).open_account_tab()

        new_password = "AnotherPassword12!"
        page = page.change_password("Test", "Teacher", new_password, password)
        assert self.is_login_page(page)
        assert is_password_updated_message_showing(self.selenium)

        page = page.login(email, new_password)

        assert self.is_dashboard_page(page)

    def test_reset_password(self):
        email, _ = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        page = self.get_to_forgotten_password_page()

        page.reset_email_submit(email)

        self.wait_for_email()

        page = email_utils.follow_reset_email_link(self.selenium, mail.outbox[0])

        new_password = "AnotherPassword12!"

        page.teacher_reset_password(new_password)

        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium).go_to_teacher_login_page().login(email, new_password)
        assert self.is_dashboard_page(page)

    def test_reset_password_fail(self):
        page = self.get_to_forgotten_password_page()
        fake_email = "fake_email@fakeemail.com"
        page.reset_email_submit(fake_email)

        time.sleep(5)

        assert len(mail.outbox) == 0

    def test_admin_sees_all_school_classes(self):
        email, password = signup_teacher_directly()
        org_name, postcode = create_organisation_directly(email)
        klass, _, access_code = create_class_directly(email, "class123")

        # create non_admin account to join the school
        # check if they cannot see classes
        non_admin_email, non_admin_password = signup_teacher_directly()
        join_teacher_to_organisation(non_admin_email, org_name, postcode, is_admin=False)

        page = (
            self.go_to_homepage()
            .go_to_teacher_login_page()
            .login(non_admin_email, non_admin_password)
            .open_classes_tab()
        )

        assert page.element_does_not_exist_by_id(f"class-code-{access_code}")

        page = self.go_to_homepage().teacher_logout()
        # then make an admin account and check
        # if the teacher can see the classes

        admin_email, admin_password = signup_teacher_directly()
        join_teacher_to_organisation(admin_email, org_name, postcode, is_admin=True)

        page = self.go_to_homepage().go_to_teacher_login_page().login(admin_email, admin_password).open_classes_tab()
        class_code_field = page.browser.find_element_by_id(f"class-code-{access_code}")
        assert class_code_field.text == access_code

    def test_admin_student_edit(self):
        email, password = signup_teacher_directly()
        org_name, postcode = create_organisation_directly(email)

        klass, _, access_code = create_class_directly(email, "class123")
        student_name, student_password, student_student = create_school_student_directly(access_code)

        joining_email, joining_password = signup_teacher_directly()
        join_teacher_to_organisation(joining_email, org_name, postcode, is_admin=True)

        page = (
            self.go_to_homepage().go_to_teacher_login_page().login(joining_email, joining_password).open_classes_tab()
        )

        class_button = WebDriverWait(self.selenium, WAIT_TIME).until(
            EC.element_to_be_clickable((By.ID, "class_button"))
        )
        class_button.click()

        edit_student_button = WebDriverWait(self.selenium, WAIT_TIME).until(
            EC.element_to_be_clickable((By.ID, "edit_student_button"))
        )
        edit_student_button.click()

        title = page.browser.find_element_by_id("student_details")
        assert title.text == f"Edit student details for {student_name} from class {klass} ({access_code})"

    def test_make_admin_popup(self):
        email, password = signup_teacher_directly()
        org_name, postcode = create_organisation_directly(email)
        page = self.go_to_homepage().go_to_teacher_login_page().login(email, password)
        joining_email, joining_password = signup_teacher_directly()

        invite_data = {"teacher_first_name": "Real", "teacher_last_name": "Name", "teacher_email": "ren@me.me"}

        for key in invite_data.keys():
            field = page.browser.find_element_by_name(key)
            field.send_keys(invite_data[key])

        invite_button = page.browser.find_element_by_name("invite_teacher_button")
        invite_button.click()

        # Once invite sent test the make admin button
        make_admin_button = WebDriverWait(self.selenium, WAIT_TIME).until(
            EC.element_to_be_clickable((By.ID, "make_admin_button_invite"))
        )
        make_admin_button.click()

        assert page.element_exists((By.CLASS_NAME, "popup-box__msg"))
        time.sleep(FADE_TIME)
        cancel_button = WebDriverWait(self.selenium, WAIT_TIME).until(
            EC.element_to_be_clickable((By.ID, "cancel_admin_popup_button"))
        )
        cancel_button.click()

        # Delete the invite and check if the form invite with
        # admin checked also makes a popup
        time.sleep(1)  # let the page scroll down
        delete_invite_button = WebDriverWait(self.selenium, WAIT_TIME).until(
            EC.element_to_be_clickable((By.ID, "delete-invite"))
        )
        delete_invite_button.click()

        for key in invite_data.keys():
            field = page.browser.find_element_by_name(key)
            field.send_keys(invite_data[key])
        checkbox = page.browser.find_element_by_name("make_admin_ticked")
        checkbox.click()

        invite_button = WebDriverWait(self.selenium, WAIT_TIME).until(
            EC.element_to_be_clickable((By.ID, "invite_teacher_button"))
        )
        invite_button.click()
        assert page.element_exists((By.CLASS_NAME, "popup-box__msg"))
        time.sleep(FADE_TIME)
        cancel_button = WebDriverWait(self.selenium, WAIT_TIME).until(
            EC.element_to_be_clickable((By.ID, "cancel_admin_popup_button"))
        )
        cancel_button.click()

        # Non admin teacher joined - make admin should also make a popup

        join_teacher_to_organisation(joining_email, org_name, postcode, is_admin=False)

        # refresh the page and scroll to the buttons
        page.browser.execute_script("location.reload()")
        page.browser.execute_script("window.scroll(0,900)")
        time.sleep(FADE_TIME)
        make_admin_button = WebDriverWait(self.selenium, WAIT_TIME).until(
            EC.element_to_be_clickable((By.ID, "make_admin_button"))
        )
        make_admin_button.click()

        assert page.element_exists((By.CLASS_NAME, "popup-box__msg"))

    def test_delete_account(self):
        FADE_TIME = 0.9  # often fails if lower

        email, password = signup_teacher_directly()
        create_organisation_directly(email)

        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium).go_to_teacher_login_page().login(email, password).open_account_tab()

        # test incorrect password
        page.browser.find_element_by_id("id_delete_password").send_keys("IncorrectPassword")
        page.browser.find_element_by_id("delete_account_button").click()
        is_message_showing(page.browser, "Your account was not deleted")

        # test cancel (no class)
        time.sleep(FADE_TIME)
        page.browser.find_element_by_id("id_delete_password").clear()
        page.browser.find_element_by_id("id_delete_password").send_keys(password)
        page.browser.find_element_by_id("delete_account_button").click()

        time.sleep(FADE_TIME)
        assert page.browser.find_element_by_id("popup-delete-review").is_displayed()
        page.browser.find_element_by_id("cancel_popup_button").click()
        time.sleep(FADE_TIME)

        # test close button in the corner
        page.browser.find_element_by_id("id_delete_password").clear()
        page.browser.find_element_by_id("id_delete_password").send_keys(password)
        page.browser.find_element_by_id("delete_account_button").click()

        time.sleep(FADE_TIME)
        page.browser.find_element_by_id("close_popup_button").click()
        time.sleep(FADE_TIME)

        # create class
        _, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        # delete then review classes
        page.browser.find_element_by_id("id_delete_password").send_keys(password)
        page.browser.find_element_by_id("delete_account_button").click()

        time.sleep(FADE_TIME)
        assert page.browser.find_element_by_id("popup-delete-review").is_displayed()
        page.browser.find_element_by_id("review_button").click()
        time.sleep(FADE_TIME)

        assert page.has_classes()
        page = page.open_account_tab()

        # test actual deletion
        page.browser.find_element_by_id("id_delete_password").send_keys(password)
        page.browser.find_element_by_id("delete_account_button").click()

        time.sleep(FADE_TIME)
        page.browser.find_element_by_id("delete_button").click()

        # back to homepage
        assert page.browser.find_element_by_class_name("banner--homepage")

        # user should not be able to login now
        page = HomePage(self.selenium).go_to_teacher_login_page().login_failure(email, password)

        assert page.has_login_failed("form-login-teacher", INVALID_LOGIN_MESSAGE)

    def test_onboarding_complete(self):
        email, password = signup_teacher_directly()

        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium).go_to_teacher_login_page().login_no_school(email, password)

        page = page.create_organisation("Test school", "W1", "GB")
        page = page.create_class("Test class", True)
        page = page.type_student_name("Test Student").create_students().complete_setup()

        assert page.has_onboarding_complete_popup()

    def get_to_forgotten_password_page(self):
        self.selenium.get(self.live_server_url)
        page = HomePage(self.selenium).go_to_teacher_login_page().go_to_teacher_forgotten_password_page()
        return page

    def wait_for_email(self):
        WebDriverWait(self.selenium, 2).until(lambda driver: len(mail.outbox) == 1)

    def is_dashboard_page(self, page):
        return page.__class__.__name__ == "TeachDashboardPage"

    def is_resources_page(self, page):
        return page.__class__.__name__ == "ResourcesPage"

    def is_onboarding_page(self, page):
        return page.__class__.__name__ == "OnboardingOrganisationPage"

    def is_login_page(self, page):
        return page.__class__.__name__ == "TeacherLoginPage"

    def is_email_verification_page(self, page):
        return page.__class__.__name__ == "EmailVerificationNeededPage"



from django.test.client import Client
from django.urls.base import reverse
from unittest import TestCase
import pytest


@pytest.mark.django_db
def test_not_logged_in():
    c = Client()
    response = c.get(reverse("reactTestSpace"))
    assert response.status_code == 403


@pytest.mark.django_db
def test_super_user():
    admin_username = "portaladmin"
    admin_password = "abc123"
    c = Client()
    c.login(username=admin_username, password=admin_password)
    response = c.get(reverse("reactTestSpace"))
    assert response.status_code == 200


@pytest.mark.django_db
def test_non_super_user():
    non_admin_username = "test teacher"
    non_admin_password = "Password1"
    c = Client()
    c.login(username=non_admin_username, password=non_admin_password)
    response = c.get(reverse("reactTestSpace"))
    assert response.status_code == 403



from aimmo.urls import HOMEPAGE_REGEX
from common.permissions import teacher_verified
from django.conf.urls import include, url
from django.http import HttpResponse
from django.views.generic import RedirectView
from django.views.generic.base import TemplateView
from django.views.i18n import JavaScriptCatalog
from game.views.level import play_default_level
from two_factor.views import BackupTokensView, ProfileView, QRGeneratorView, SetupCompleteView

from portal.helpers.decorators import ratelimit
from portal.helpers.ratelimit import RATELIMIT_LOGIN_GROUP, RATELIMIT_METHOD, RATELIMIT_LOGIN_RATE
from portal.helpers.regexes import ACCESS_CODE_REGEX
from portal.reactTestSpace import reactTestSpace
from portal.views.about import about, getinvolved, contribute
from portal.views.admin import AdminChangePasswordDoneView, AdminChangePasswordView
from portal.views.aimmo.dashboard import StudentAimmoDashboard, TeacherAimmoDashboard
from portal.views.api import (
    AnonymiseOrphanSchoolsView,
    InactiveUsersView,
    last_connected_since,
    number_users_per_country,
    registered_users,
)
from portal.views.dotmailer import dotmailer_consent_form, process_newsletter_form
from portal.views.email import send_new_users_report, verify_email
from portal.views.home import (
    coding_club,
    download_student_pack,
    home,
    home_learning,
    logout_view,
    register_view,
    reset_screentime_warning,
)
from portal.views.legal import privacy_policy, terms
from portal.views.login import old_login_form_redirect
from portal.views.login.independent_student import IndependentStudentLoginView
from portal.views.login.student import StudentLoginView, StudentClassCodeView, student_direct_login
from portal.views.login.teacher import TeacherLoginView
from portal.views.organisation import organisation_leave, organisation_manage
from portal.views.play_landing_page import play_landing_page
from portal.views.registration import (
    password_reset_check_and_confirm,
    password_reset_done,
    student_password_reset,
    teacher_password_reset,
    delete_account,
)
from portal.views.student.edit_account_details import (
    independentStudentEditAccountView,
    SchoolStudentEditAccountView,
    student_edit_account,
)
from portal.views.student.play import SchoolStudentDashboard, IndependentStudentDashboard, student_join_organisation
from portal.views.teach import teach
from portal.views.teacher.dashboard import (
    dashboard_manage,
    organisation_kick,
    invite_toggle_admin,
    organisation_toggle_admin,
    teacher_accept_student_request,
    teacher_disable_2FA,
    teacher_reject_student_request,
    delete_teacher_invite,
    invited_teacher,
    resend_invite_teacher,
)
from portal.views.teacher.teach import (
    teacher_class_password_reset,
    teacher_delete_class,
    teacher_delete_students,
    teacher_dismiss_students,
    teacher_edit_class,
    teacher_edit_student,
    teacher_move_students,
    teacher_move_students_to_class,
    teacher_onboarding_create_class,
    teacher_onboarding_edit_class,
    teacher_print_reminder_cards,
    teacher_download_csv,
    teacher_view_class,
)
from portal.views.teacher.teacher_resources import (
    teacher_kurono_resources,
    teacher_rapid_router_resources,
    kurono_teaching_packs,
    materials,
)
from portal.views.two_factor.core import CustomSetupView
from portal.views.two_factor.profile import CustomDisableView

js_info_dict = {"packages": ("conf.locale",)}

two_factor_patterns = [
    url(r"^account/two_factor/setup/$", CustomSetupView.as_view(), name="setup"),
    url(r"^account/two_factor/qrcode/$", QRGeneratorView.as_view(), name="qr"),
    url(r"^account/two_factor/setup/complete/$", SetupCompleteView.as_view(), name="setup_complete"),
    url(r"^account/two_factor/backup/tokens/$", teacher_verified(BackupTokensView.as_view()), name="backup_tokens"),
    url(r"^account/two_factor/$", teacher_verified(ProfileView.as_view()), name="profile"),
    url(r"^account/two_factor/disable/$", teacher_verified(CustomDisableView.as_view()), name="disable"),
]


urlpatterns = [
    url(HOMEPAGE_REGEX, include("aimmo.urls")),
    url(r"^teach/kurono/dashboard/$", TeacherAimmoDashboard.as_view(), name="teacher_aimmo_dashboard"),
    url(r"^play/kurono/dashboard/$", StudentAimmoDashboard.as_view(), name="student_aimmo_dashboard"),
    url(r"^favicon\.ico$", RedirectView.as_view(url="/static/portal/img/favicon.ico", permanent=True)),
    url(r"^administration/password_change/$", AdminChangePasswordView.as_view(), name="administration_password_change"),
    url(
        r"^administration/password_change_done/$",
        AdminChangePasswordDoneView.as_view(),
        name="administration_password_change_done",
    ),
    url(r"^mail/weekly", send_new_users_report, name="send_new_users_report"),
    url(r"^users/inactive/", InactiveUsersView.as_view(), name="inactive_users"),
    url(r"^locked_out/$", TemplateView.as_view(template_name="portal/locked_out.html"), name="locked_out"),
    url(r"^", include((two_factor_patterns, "two_factor"), namespace="two_factor")),
    url(r"^i18n/", include("django.conf.urls.i18n")),
    url(r"^jsi18n/$", JavaScriptCatalog.as_view(), js_info_dict),
    url(r"^(?P<levelName>[A-Z0-9]+)/$", play_default_level, name="play_default_level"),
    url(r"^$", home, name="home"),
    url(r"^home-learning", home_learning, name="home-learning"),
    url(r"^register_form", register_view, name="register"),
    url(
        r"^login/teacher/$",
        # The ratelimit decorator checks how often a POST request is performed on that view.
        # It checks against the username value specifically. If the number of requests
        # exceeds the specified rate, then the user will be blocked (if block = True).
        ratelimit(
            group=RATELIMIT_LOGIN_GROUP,
            key="post:auth-username",
            method=RATELIMIT_METHOD,
            rate=RATELIMIT_LOGIN_RATE,
            block=True,
        )(TeacherLoginView.as_view()),
        name="teacher_login",
    ),
    url(
        rf"^login/student/(?P<access_code>{ACCESS_CODE_REGEX})/(?:(?P<login_type>classform)/)?$",
        StudentLoginView.as_view(),
        name="student_login",
    ),
    url(r"^login/student/$", StudentClassCodeView.as_view(), name="student_login_access_code"),
    url(r"^u/(?P<user_id>[0-9]+)/(?P<login_id>[a-z0-9]+)/$", student_direct_login, name="student_direct_login"),
    url(
        r"^login/independent/$",
        ratelimit(
            group=RATELIMIT_LOGIN_GROUP,
            key="post:username",
            method=RATELIMIT_METHOD,
            rate=RATELIMIT_LOGIN_RATE,
            block=True,
            is_teacher=False,
        )(IndependentStudentLoginView.as_view()),
        name="independent_student_login",
    ),
    url(r"^login_form", old_login_form_redirect, name="old_login_form"),
    url(r"^logout/$", logout_view, name="logout_view"),
    url(r"^news_signup/$", process_newsletter_form, name="process_newsletter_form"),
    url(r"^consent_form/$", dotmailer_consent_form, name="consent_form"),
    url(
        r"^verify_email/$",
        TemplateView.as_view(template_name="portal/email_verification_needed.html"),
        name="email_verification",
    ),
    url(r"^verify_email/(?P<token>[0-9a-f]+)/$", verify_email, name="verify_email"),
    url(r"^user/password/reset/student/$", student_password_reset, name="student_password_reset"),
    url(r"^user/password/reset/teacher/$", teacher_password_reset, name="teacher_password_reset"),
    url(r"^user/password/reset/done/$", password_reset_done, name="reset_password_email_sent"),
    url(
        r"^user/password/reset/(?P<uidb64>[0-9A-Za-z]+)-(?P<token>.+)/$",
        password_reset_check_and_confirm,
        name="password_reset_check_and_confirm",
    ),
    url(r"^user/reset_screentime_warning/$", reset_screentime_warning, name="reset_screentime_warning"),
    url(r"^user/reset_session_time/$", lambda _: HttpResponse(status=204), name="reset_session_time"),
    url(
        r"^teacher/password/reset/complete/$",
        TemplateView.as_view(template_name="portal/reset_password_done.html"),
        name="password_reset_complete",
    ),
    url(r"^teach/$", teach, name="teach"),
    url(r"^teach/onboarding-organisation/$", organisation_manage, name="onboarding-organisation"),
    url(r"^teach/onboarding-classes", teacher_onboarding_create_class, name="onboarding-classes"),
    url(
        rf"^teach/onboarding-class/(?P<access_code>{ACCESS_CODE_REGEX})$",
        teacher_onboarding_edit_class,
        name="onboarding-class",
    ),
    url(
        rf"^teach/onboarding-class/(?P<access_code>{ACCESS_CODE_REGEX})/print_reminder_cards/$",
        teacher_print_reminder_cards,
        name="teacher_print_reminder_cards",
    ),
    url(
        rf"^teach/onboarding-class/(?P<access_code>{ACCESS_CODE_REGEX})/download_csv/$",
        teacher_download_csv,
        name="teacher_download_csv",
    ),
    url(r"^invited_teacher/(?P<token>[0-9a-f]+)/$", invited_teacher, name="invited_teacher"),
    url(r"^play/$", play_landing_page, name="play"),
    url(r"^play/details/$", SchoolStudentDashboard.as_view(), name="student_details"),
    url(r"^play/details/independent$", IndependentStudentDashboard.as_view(), name="independent_student_details"),
    url(r"^play/account/$", student_edit_account, name="student_edit_account"),
    url(
        r"^play/account/independent/$",
        ratelimit(
            group=RATELIMIT_LOGIN_GROUP,
            key="post:name",
            method=RATELIMIT_METHOD,
            rate=RATELIMIT_LOGIN_RATE,
            block=True,
            is_teacher=False,
        )(independentStudentEditAccountView),
        name="independent_edit_account",
    ),
    url(r"^play/account/school_student/$", SchoolStudentEditAccountView.as_view(), name="school_student_edit_account"),
    url(r"^play/join/$", student_join_organisation, name="student_join_organisation"),
    url(r"^about", about, name="about"),
    url(r"^getinvolved", getinvolved, name="getinvolved"),
    url(r"^contribute", contribute, name="contribute"),
    url(r"^terms", terms, name="terms"),
    url(r"^privacy-policy/$", privacy_policy, name="privacy_policy"),
    url(r"^teach/materials/$", materials, name="materials"),
    url(r"^teach/kurono_teaching_packs$", kurono_teaching_packs, name="kurono_packs"),
    url(r"^teach/resources/$", teacher_rapid_router_resources, name="teaching_resources"),
    url(r"^teach/kurono_resources/$", teacher_kurono_resources, name="kurono_teaching_resources"),
    url(r"^teach/dashboard/$", dashboard_manage, name="dashboard"),
    url(r"^teach/dashboard/kick/(?P<pk>[0-9]+)/$", organisation_kick, name="organisation_kick"),
    url(r"^teach/dashboard/toggle_admin/(?P<pk>[0-9]+)/$", organisation_toggle_admin, name="organisation_toggle_admin"),
    url(r"^teach/dashboard/disable_2FA/(?P<pk>[0-9]+)/$", teacher_disable_2FA, name="teacher_disable_2FA"),
    url(r"^teach/dashboard/school/leave/$", organisation_leave, name="organisation_leave"),
    url(
        r"^teach/dashboard/student/accept/(?P<pk>[0-9]+)/$",
        teacher_accept_student_request,
        name="teacher_accept_student_request",
    ),
    url(
        r"^teach/dashboard/student/reject/(?P<pk>[0-9]+)/$",
        teacher_reject_student_request,
        name="teacher_reject_student_request",
    ),
    url(rf"^teach/class/(?P<access_code>{ACCESS_CODE_REGEX})$", teacher_view_class, name="view_class"),
    url(
        rf"^teach/class/delete/(?P<access_code>{ACCESS_CODE_REGEX})$", teacher_delete_class, name="teacher_delete_class"
    ),
    url(
        rf"^teach/class/(?P<access_code>{ACCESS_CODE_REGEX})/students/delete/$",
        teacher_delete_students,
        name="teacher_delete_students",
    ),
    url(rf"^teach/class/edit/(?P<access_code>{ACCESS_CODE_REGEX})$", teacher_edit_class, name="teacher_edit_class"),
    url(r"^teach/class/student/edit/(?P<pk>[0-9]+)/$", teacher_edit_student, name="teacher_edit_student"),
    url(
        rf"^teach/class/(?P<access_code>{ACCESS_CODE_REGEX})/password_reset/$",
        teacher_class_password_reset,
        name="teacher_class_password_reset",
    ),
    url(
        rf"^teach/class/(?P<access_code>{ACCESS_CODE_REGEX})/students/dismiss/$",
        teacher_dismiss_students,
        name="teacher_dismiss_students",
    ),
    url(
        rf"^teach/class/(?P<access_code>{ACCESS_CODE_REGEX})/students/move/$",
        teacher_move_students,
        name="teacher_move_students",
    ),
    url(r"^teach/dashboard/resend_invite/(?P<token>[0-9a-f]+)/$", resend_invite_teacher, name="resend_invite_teacher"),
    url(
        r"^teach/dashboard/toggle_admin_invite/(?P<invite_id>[0-9]+)/$", invite_toggle_admin, name="invite_toggle_admin"
    ),
    url(
        r"^teach/dashboard/delete_teacher_invite/(?P<token>[0-9a-f]+)$",
        delete_teacher_invite,
        name="delete_teacher_invite",
    ),
    url(
        rf"^teach/class/(?P<access_code>{ACCESS_CODE_REGEX})/students/move/disambiguate/$",
        teacher_move_students_to_class,
        name="teacher_move_students_to_class",
    ),
    url(r"^delete/account/$", delete_account, name="delete_account"),
    url(
        r"^schools/anonymise/(?P<start_id>\d+)/", AnonymiseOrphanSchoolsView.as_view(), name="anonymise_orphan_schools"
    ),
    url(
        r"^api/",
        include(
            [
                url(
                    r"^registered/(?P<year>\d{4})/(?P<month>\d{2})/(?P<day>\d{2})/$",
                    registered_users,
                    name="registered-users",
                ),
                url(
                    r"^lastconnectedsince/(?P<year>\d{4})/(?P<month>\d{2})/(?P<day>\d{2})/$",
                    last_connected_since,
                    name="last-connected-since",
                ),
                url(
                    r"^userspercountry/(?P<country>(AF|AX|AL|DZ|AS|AD|AO|AI|AQ|AG|AR|AM|AW|AU|AT|AZ|BS|BH|BD|BB|BY|BE|BZ|BJ|BM|BT|BO|BQ|BA|BW|BV|BR|IO|BN|BG|BF|BI|KH|CM|CA|CV|KY|CF|TD|CL|CN|CX|CC|CO|KM|CG|CD|CK|CR|CI|HR|CU|CW|CY|CZ|DK|DJ|DM|DO|EC|EG|SV|GQ|ER|EE|ET|FK|FO|FJ|FI|FR|GF|PF|TF|GA|GM|GE|DE|GH|GI|GR|GL|GD|GP|GU|GT|GG|GN|GW|GY|HT|HM|VA|HN|HK|HU|IS|IN|ID|IR|IQ|IE|IM|IL|IT|JM|JP|JE|JO|KZ|KE|KI|KP|KR|KW|KG|LA|LV|LB|LS|LR|LY|LI|LT|LU|MO|MK|MG|MW|MY|MV|ML|MT|MH|MQ|MR|MU|YT|MX|FM|MD|MC|MN|ME|MS|MA|MZ|MM|NA|NR|NP|NL|NC|NZ|NI|NE|NG|NU|NF|MP|NO|OM|PK|PW|PS|PA|PG|PY|PE|PH|PN|PL|PT|PR|QA|RE|RO|RU|RW|BL|SH|KN|LC|MF|PM|VC|WS|SM|ST|SA|SN|RS|SC|SL|SG|SX|SK|SI|SB|SO|ZA|GS|SS|ES|LK|SD|SR|SJ|SZ|SE|CH|SY|TW|TJ|TZ|TH|TL|TG|TK|TO|TT|TN|TR|TM|TC|TV|UG|UA|AE|GB|US|UM|UY|UZ|VU|VE|VN|VG|VI|WF|EH|YE|ZM|ZW))/$",
                    number_users_per_country,
                    name="number_users_per_country",
                ),
            ]
        ),
    ),
    url(r"^reactTestSpace/", reactTestSpace, name="reactTestSpace"),
    url(r"^codingClub/$", coding_club, name="codingClub"),
    url(r"^codingClub/(?P<student_pack_type>[3-4])/", download_student_pack, name="download_student_pack"),
]



from aimmo.templatetags.players_utils import get_user_playable_games
from aimmo.worksheets import get_complete_worksheets, get_incomplete_worksheets
from common import app_settings as common_app_settings
from common.models import EmailVerification
from common.permissions import logged_in_as_teacher
from common.utils import using_two_factor
from django import template
from django.conf import settings
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.shortcuts import reverse
from django.template.context import RequestContext
from django.template.defaultfilters import stringfilter
from portal import __version__, beta

register = template.Library()


@register.filter(name="emaildomain")
@stringfilter
def emaildomain(email):
    return "*********" + email[email.find("@") :]


@register.filter(name="has_2FA")
def has_2FA(u):
    return using_two_factor(u)


@register.filter(name="is_logged_in")
def is_logged_in(u):
    return (
        u
        and u.is_authenticated
        and (not using_two_factor(u) or (hasattr(u, "is_verified") and u.is_verified()))
    )


@register.filter
def is_verified(u: User) -> bool:
    try:
        verifications = EmailVerification.objects.filter(user=u)
        latest_verification = verifications.latest("user")
    except ObjectDoesNotExist:
        return False

    return latest_verification.verified


@register.filter
def is_developer(u):
    return not u.is_anonymous and u.userprofile.developer


@register.filter
def is_production(request):
    # 'production' excludes localhost/dev/staging
    return common_app_settings.MODULE_NAME == "default"


@register.filter
def has_beta_access(request):
    return beta.has_beta_access(request)


@register.inclusion_tag("portal/partials/aimmo_games_table.html", takes_context=True)
def games_table(context, base_url):
    playable_games = get_user_playable_games(context, base_url)

    playable_games["complete_worksheets"] = get_complete_worksheets()
    playable_games["incomplete_worksheets"] = get_incomplete_worksheets()

    return playable_games


@register.filter(name="make_into_username")
def make_into_username(u):
    username = ""
    if hasattr(u, "userprofile"):
        if hasattr(u.userprofile, "student"):
            username = u.first_name
        elif hasattr(u.userprofile, "teacher"):
            username = u.first_name + " " + u.last_name

    return username


@register.filter(name="is_logged_in_as_teacher")
def is_logged_in_as_teacher(u):
    return is_logged_in(u) and u.userprofile and hasattr(u.userprofile, "teacher")


@register.filter(name="is_logged_in_as_student")
def is_logged_in_as_student(u):
    return (
        is_logged_in(u)
        and u.userprofile
        and hasattr(u.userprofile, "student")
        and u.userprofile.student.class_field != None
    )


@register.filter(name="is_independent_student")
def is_independent_student(u):
    return (
        is_logged_in(u)
        and u.userprofile
        and hasattr(u.userprofile, "student")
        and u.userprofile.student.is_independent()
    )


@register.filter(name="has_teacher_finished_onboarding")
def has_teacher_finished_onboarding(u):
    teacher = u.userprofile.teacher
    return (
        is_logged_in_as_teacher(u)
        and teacher.has_school()
    )


@register.filter(name="is_logged_in_as_school_user")
def is_logged_in_as_school_user(u):
    return (
        is_logged_in(u)
        and u.userprofile
        and (
            (
                hasattr(u.userprofile, "student")
                and u.userprofile.student.class_field != None
            )
            or hasattr(u.userprofile, "teacher")
        )
    )


@register.filter(name="get_user_status")
def get_user_status(u):
    if is_logged_in_as_school_user(u):
        if is_logged_in_as_teacher(u):
            return "TEACHER"
        else:
            return "SCHOOL_STUDENT"
    elif is_logged_in(u):
        return "INDEPENDENT_STUDENT"
    else:
        return "UNTRACKED"


@register.filter(name="make_title_caps")
def make_title_caps(s):
    if len(s) > 0:
        s = s[0].upper() + s[1:]
    return s


@register.filter(name="cloud_storage")
@stringfilter
def cloud_storage(e):
    return settings.CLOUD_STORAGE_PREFIX + e


@register.filter(name="get_project_version")
def get_project_version():
    return __version__


@register.simple_tag(takes_context=True)
def url_for_aimmo_dashboard(context: RequestContext):
    if logged_in_as_teacher(context.request.user):
        return reverse("teacher_aimmo_dashboard")
    else:
        return reverse("student_aimmo_dashboard")



from two_factor.forms import (
    TOTPDeviceForm,
    MethodForm,
    DeviceValidationForm,
    YubiKeyDeviceForm,
)
from two_factor.views.core import SetupView


# This custom class gets rid of the 'welcome' step of 2FA
# which the new design not needs any more
class CustomSetupView(SetupView):
    form_list = (
        ("generator", TOTPDeviceForm),
        ("method", MethodForm),
        ("validation", DeviceValidationForm),
        ("yubikey", YubiKeyDeviceForm),
    )

    condition_dict = {
        "call": lambda self: self.get_method() == "call",
        "sms": lambda self: self.get_method() == "sms",
        "validation": lambda self: self.get_method() in ("sms", "call"),
        "yubikey": lambda self: self.get_method() == "yubikey",
    }



# -*- coding: utf-8 -*-
import re
import sys

from setuptools import find_packages, setup

with open("portal/__init__.py", "r") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

try:
    from semantic_release import setup_hook

    setup_hook(sys.argv)
except ImportError:
    pass

setup(
    name="codeforlife-portal",
    version=version,
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "django==3.2.15",
        "django-countries==7.3.1",
        "djangorestframework==3.13.1",
        "django-pipeline==2.0.8",
        "django-recaptcha==2.0.6",
        "pyyaml==5.4.1",
        "rapid-router>=4",
        "aimmo>=2",
        "reportlab==3.6.1",
        "django-formtools==2.2",
        "django-otp==1.0.2",  # we needed to fix this due to a wide ranged dependency in django-two-factor-auth
        "requests==2.25.0",
        "django-treebeard==4.3.1",
        "django-sekizai==2.0.0",
        "django-classy-tags==2.0.0",
        "sqlparse==0.4.2",
        "libsass==0.21.0",
        "phonenumbers==8.12.12",
        "more-itertools==8.6.0",
        "future==0.18.2",
        f"cfl-common=={version}",
        "django-ratelimit==3.0.1",
        "django-preventconcurrentlogins==0.8.2",
        "django-csp==3.7",
        "setuptools==62.1.0",
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Framework :: Django",
    ],
    zip_safe=False,
)



from django import template
from common.models import AimmoCharacter

register = template.Library()


@register.inclusion_tag("portal/partials/character_list.html")
def character_list():
    """
    Registers the inclusion tag for the character card list partial.
    The template currently expects a list of elements which each contain the following:
    - name: the heading of the card
    - image_path: the path to the card's image
    - description: the text paragraph of the card
    """
    return {"characters": AimmoCharacter.objects.sorted()}



from __future__ import absolute_import

from common.models import Class, Teacher
from common.tests.utils.classes import create_class_directly
from common.tests.utils.organisation import (
    create_organisation_directly,
    join_teacher_to_organisation,
)
from common.tests.utils.student import create_school_student_directly
from common.tests.utils.teacher import signup_teacher_directly
from django.test import Client, TestCase
from django.urls import reverse

from .pageObjects.portal.teach.class_page import TeachClassPage
from .base_test import BaseTest
from .utils.classes import create_class
from .utils.messages import is_class_created_message_showing


class TestClass(TestCase):
    def test_delete_class(self):
        email1, password1 = signup_teacher_directly()
        email2, password2 = signup_teacher_directly()
        create_organisation_directly(email1)
        klass, klass_name, access_code = create_class_directly(email1)
        _, _, student = create_school_student_directly(access_code)

        c = Client()

        url = reverse("teacher_delete_class", kwargs={"access_code": access_code})

        # Login as another teacher, try to delete the class and check for 404
        c.login(username=email2, password=password2)

        response = c.post(url)

        assert response.status_code == 404

        c.logout()

        # Login as first teacher, check there is a class
        c.login(username=email1, password=password1)

        teacher = Teacher.objects.get(new_user__username=email1)
        teacher_classes = Class.objects.filter(teacher=teacher)

        assert len(teacher_classes) == 1

        # Try to delete the class, check that it can't be deleted since it's not empty
        response = c.post(url)

        teacher_classes = Class.objects.filter(teacher=teacher)

        assert response.status_code == 302
        assert len(teacher_classes) == 1
        assert teacher.has_class() == True

        # Delete the student, and try again, check the class is deleted successfully
        student.delete()

        response = c.post(url)

        teacher_classes = Class.objects.filter(teacher=teacher)

        assert response.status_code == 302
        assert len(teacher_classes) == 0
        assert teacher.has_class() == False

        # Check class is anonymised
        new_klass = Class._base_manager.get(pk=klass.id)
        assert new_klass.name != klass_name
        assert new_klass.access_code == ""
        assert not new_klass.is_active

    def test_edit_class(self):
        email1, password1 = signup_teacher_directly()
        email2, password2 = signup_teacher_directly()
        create_organisation_directly(email1)
        _, class_name, access_code = create_class_directly(email1)
        create_school_student_directly(access_code)

        c = Client()

        url = reverse("teacher_edit_class", kwargs={"access_code": access_code})
        new_class_name = "New class name"
        data = {
            "name": new_class_name,
            "classmate_progress": "on",
            "external_requests": "1000",  # Setting to always accept requests
            "class_edit_submit": "",
        }

        # Login as another teacher, try to edit the class and check for 404
        c.login(username=email2, password=password2)

        response = c.post(url, data)

        assert response.status_code == 404

        c.logout()

        # Login as first teacher, check the default class settings
        c.login(username=email1, password=password1)

        teacher = Teacher.objects.get(new_user__username=email1)
        klass = Class.objects.get(teacher=teacher)

        assert klass.name == class_name
        assert not klass.classmates_data_viewable
        assert not klass.accept_requests_until

        # Edit class settings, check they match the data dict above
        response = c.post(url, data)

        assert response.status_code == 302

        klass = Class.objects.get(teacher=teacher)

        assert klass.name == new_class_name
        assert klass.classmates_data_viewable
        assert klass.always_accept_requests

    def test_transfer_class(self):
        email1, password1 = signup_teacher_directly()
        email2, password2 = signup_teacher_directly()
        school_name, postcode = create_organisation_directly(email1)
        join_teacher_to_organisation(email2, school_name, postcode)
        klass1, _, access_code1 = create_class_directly(email1)
        klass2, _, access_code2 = create_class_directly(email2)
        _, _, student1 = create_school_student_directly(access_code1)
        _, _, student2 = create_school_student_directly(access_code2)

        teacher1 = Teacher.objects.get(new_user__username=email1)
        teacher2 = Teacher.objects.get(new_user__username=email2)
        teacher1_classes = Class.objects.filter(teacher=teacher1)
        teacher2_classes = Class.objects.filter(teacher=teacher2)

        assert len(teacher1_classes) == 1
        assert len(teacher2_classes) == 1

        c = Client()

        url = reverse("teacher_edit_class", kwargs={"access_code": access_code1})
        data = {"new_teacher": teacher2.id, "class_move_submit": ""}

        # Login as first teacher and transfer class to the second teacher
        c.login(username=email1, password=password1)

        response = c.post(url, data)

        assert response.status_code == 302

        c.logout()

        teacher1_classes = Class.objects.filter(teacher=teacher1)
        teacher2_classes = Class.objects.filter(teacher=teacher2)

        assert len(teacher1_classes) == 0
        assert len(teacher2_classes) == 2
        assert teacher2_classes[1] == klass2
        assert not teacher1.teaches(student2.user)
        assert teacher2.teaches(student2.user)


# Class for Selenium tests. We plan to replace these and turn them into Cypress tests
class TestClassFrontend(BaseTest):
    def test_create(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        page = self.go_to_homepage().go_to_teacher_login_page().login_no_class(email, password).open_classes_tab()

        assert page.does_not_have_classes()

        page, class_name = create_class(page)
        assert is_class_created_message_showing(self.selenium, class_name)

    def test_create_class_as_admin_for_another_teacher(self):
        email1, password1 = signup_teacher_directly()
        email2, password2 = signup_teacher_directly()
        teacher2 = Teacher.objects.get(new_user__email=email2)
        name, postcode = create_organisation_directly(email1)
        join_teacher_to_organisation(email2, name, postcode)

        # Check teacher 2 doesn't have any classes
        page = self.go_to_homepage().go_to_teacher_login_page().login(email2, password2).open_classes_tab()
        assert page.does_not_have_classes()
        page.logout()

        # Log in as the first teacher and create a class for the second one
        page = self.go_to_homepage().go_to_teacher_login_page().login(email1, password1).open_classes_tab()
        page, class_name = create_class(page, teacher_id=teacher2.id)
        page = TeachClassPage(page.browser)
        assert is_class_created_message_showing(self.selenium, class_name)
        page.logout()

        # Check teacher 2 now has the class
        page = self.go_to_homepage().go_to_teacher_login_page().login(email2, password2).open_classes_tab()
        assert page.has_classes()

    def test_create_dashboard(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        klass, name, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        page = self.go_to_homepage().go_to_teacher_login_page().login(email, password).open_classes_tab()

        page, class_name = create_class(page)

        assert is_class_created_message_showing(self.selenium, class_name)

    def test_create_dashboard_non_admin(self):
        email_1, password_1 = signup_teacher_directly()
        email_2, password_2 = signup_teacher_directly()
        name, postcode = create_organisation_directly(email_1)
        klass_1, class_name_1, access_code_1 = create_class_directly(email_1)
        create_school_student_directly(access_code_1)
        join_teacher_to_organisation(email_2, name, postcode)
        klass_2, class_name_2, access_code_2 = create_class_directly(email_2)
        create_school_student_directly(access_code_2)

        page = self.go_to_homepage().go_to_teacher_login_page().login(email_2, password_2).open_classes_tab()

        page, class_name_3 = create_class(page)

        assert is_class_created_message_showing(self.selenium, class_name_3)



from __future__ import absolute_import

from .base_page import BasePage


class PasswordResetPage(BasePage):
    def __init__(self, browser):
        super(PasswordResetPage, self).__init__(browser)

        assert self.on_correct_page("reset_password_form_page")

    def teacher_reset_password(self, new_password):
        self.clear_and_fill("new_password1", new_password)
        self.clear_and_fill("new_password2", new_password)

        self.browser.find_element_by_id("teacher_update_button").click()

        self.wait_for_element_by_id("reset_password_done_page")

        return self

    def student_reset_password(self, new_password):
        self.clear_and_fill("new_password1", new_password)
        self.clear_and_fill("new_password2", new_password)

        self.browser.find_element_by_id("student_update_button").click()

        self.wait_for_element_by_id("reset_password_done_page")

        return self

    def clear_and_fill(self, field, value):
        self.browser.find_element_by_id("id_" + field).clear()
        self.browser.find_element_by_id("id_" + field).send_keys(value)



from typing import Any, Dict, Optional, List

from aimmo.models import Game
from common import email_messages
from common.helpers.emails import NOTIFICATION_EMAIL, send_email
from common.models import Student
from common.permissions import (
    logged_in_as_independent_student,
    logged_in_as_school_student,
)
from common.utils import LoginRequiredNoErrorMixin
from django.contrib import messages
from django.contrib.auth.decorators import login_required, user_passes_test
from django.contrib.auth.mixins import UserPassesTestMixin
from django.db.models.query import QuerySet
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse_lazy
from django.views.generic.base import TemplateView
from game.models import Level, Attempt

from portal.forms.play import StudentJoinOrganisationForm

from django.views.generic.edit import FormView


class SchoolStudentDashboard(LoginRequiredNoErrorMixin, UserPassesTestMixin, TemplateView):
    template_name = "portal/play/student_dashboard.html"
    login_url = reverse_lazy("student_login_access_code")

    def test_func(self) -> Optional[bool]:
        return logged_in_as_school_student(self.request.user)

    def get_context_data(self, **kwargs: Any) -> Dict[str, Any]:
        """
        Gathers the context data required by the template. First, the student's scores
        for the original Rapid Router levels is gathered, second, the student's scores
        for any levels shared with them by their teacher, and third, the student's
        Kurono game information if they have one.
        """
        # Get score data for all original levels
        levels = Level.objects.sorted_levels()
        student = self.request.user.new_student

        context_data = _compute_rapid_router_scores(student, levels)

        # Find any custom levels created by the teacher and shared with the student
        klass = student.class_field
        teacher = klass.teacher.user
        custom_levels = student.new_user.shared.filter(owner=teacher)

        if custom_levels:
            custom_levels_data = _compute_rapid_router_scores(student, custom_levels)

            context_data["total_custom_score"] = custom_levels_data["total_score"]
            context_data["total_custom_available_score"] = custom_levels_data["total_available_score"]

        # Get Kurono game info if the class has a game linked to it
        aimmo_game = klass.active_game
        if aimmo_game:
            active_worksheet = aimmo_game.worksheet

            context_data["worksheet_id"] = active_worksheet.id
            context_data["worksheet_image"] = active_worksheet.image_path

        return context_data


class IndependentStudentDashboard(LoginRequiredNoErrorMixin, UserPassesTestMixin, TemplateView, FormView):
    template_name = "portal/play/independent_student_dashboard.html"
    login_url = reverse_lazy("independent_student_login")

    def test_func(self) -> Optional[bool]:
        return logged_in_as_independent_student(self.request.user)

    def get_context_data(self, **kwargs: Any) -> Dict[str, Any]:
        levels = Level.objects.sorted_levels()
        student = self.request.user.new_student

        return _compute_rapid_router_scores(
            student,
            levels,
        )


def _compute_rapid_router_scores(student: Student, levels: List[Level] or QuerySet) -> Dict[str, int]:
    """
    Finds Rapid Router progress and score data for a specific student and a specific
    set of levels. This is used to show quick score data to the student on their
    dashboard.
    :param student: the student whose progress this is looking for
    :param levels: the list of levels to gather the progress data of
    :return: a dictionary of integers:
    - num_completed: number of completed levels. A completed level is a level that has a
    successful attempt (van made it to the final house) regardless of the final score.
    - num_top_scores: number of levels that have been completed with a full final score
    of either 10/10 or 20/20 (depending on whether the level has route score enabled)
    - total_score: the addition of all the completed levels' final scores
    - total_available_score: the addition of the maximum attainable score of all levels
    """
    num_completed = num_top_scores = total_available_score = 0
    total_score = 0.0
    # Get a QuerySet of best attempts for each level
    best_attempts = Attempt.objects.filter(level__in=levels, student=student, is_best_attempt=True).select_related(
        "level"
    )

    # Calculate total available score. A level has a max score of 20 by default unless
    # its route score is disabled or it is a custom level (not in an episode)
    for level in levels:
        max_score = 10 if level.disable_route_score or not level.episode else 20
        total_available_score += max_score

    # For each level, compare best attempt's score with level's max score and increment
    # variables as needed
    if best_attempts:
        attempts_dict = {best_attempt.level.id: best_attempt for best_attempt in best_attempts}
        for level in levels:
            max_score = 10 if level.disable_route_score or not level.episode else 20
            attempt = attempts_dict.get(level.id)

            if attempt and attempt.score:
                num_completed += 1
                if attempt.score == max_score:
                    num_top_scores += 1

                total_score += attempt.score

    return {
        "num_completed": num_completed,
        "num_top_scores": num_top_scores,
        "total_score": int(total_score),
        "total_available_score": total_available_score,
    }


def username_labeller(request):
    return request.user.username


@login_required(login_url=reverse_lazy("independent_student_login"))
@user_passes_test(
    logged_in_as_independent_student,
    login_url=reverse_lazy("independent_student_login"),
)
def student_join_organisation(request):

    student = request.user.new_student
    request_form = StudentJoinOrganisationForm()

    # check student not managed by a school
    if student.class_field:
        return HttpResponseRedirect(reverse_lazy("student_details"))

    if request.method == "POST":
        if "class_join_request" in request.POST:
            request_form = StudentJoinOrganisationForm(request.POST)
            process_join_organisation_form(request_form, request, student)

        elif "revoke_join_request" in request.POST:
            student.pending_class_request = None
            student.save()
            # Check teacher hasn't since accepted rejection before posting success
            show_cancellation_message_if_student_not_in_class(student, request)
            return HttpResponseRedirect(reverse_lazy("student_edit_account"))

    res = render(
        request,
        "portal/play/student_join_organisation.html",
        {"request_form": request_form, "student": student},
    )
    return res


def process_join_organisation_form(request_form, request, student):
    if request_form.is_valid():
        student.pending_class_request = request_form.klass
        student.save()

        email_message = email_messages.studentJoinRequestSentEmail(
            request,
            request_form.klass.teacher.school.name,
            request_form.klass.access_code,
        )
        send_email(
            NOTIFICATION_EMAIL,
            [student.new_user.email],
            email_message["subject"],
            email_message["message"],
            email_message["subject"],
        )

        email_message = email_messages.studentJoinRequestNotifyEmail(
            request,
            student.new_user.username,
            student.new_user.email,
            student.pending_class_request.access_code,
        )
        send_email(
            NOTIFICATION_EMAIL,
            [student.pending_class_request.teacher.new_user.email],
            email_message["subject"],
            email_message["message"],
            email_message["subject"],
        )

        messages.success(request, "Your request to join a school has been received successfully.")


def show_cancellation_message_if_student_not_in_class(student, request):
    if not student.class_field:
        messages.success(request, "Your request to join a school has been cancelled successfully.")



from aimmo.models import Game
from common.models import Class


def generate_name():
    name = "Game %d" % generate_name.next_id

    generate_name.next_id += 1

    return name


generate_name.next_id = 1


def create_aimmo_game_directly(klass: Class, worksheet_id: int) -> Game:
    """Generate an aimmo game with the details given.

    Args:
        klass (Class): The instance of the class.
        worksheet_id (int): The id of the worksheet.

    Returns:
        game: Game: The game model instance.
    """
    name = generate_name()

    game = Game.objects.create(name=name, game_class=klass, worksheet_id=worksheet_id)

    return game



from __future__ import absolute_import

from .base_page import BasePage


class KuronoTeacherDashboardPage(BasePage):
    def __init__(self, browser):
        super(KuronoTeacherDashboardPage, self).__init__(browser)

        assert self.on_correct_page("kurono_teacher_dashboard_page")

    def create_game(self, class_id):
        self._click_add_game_dropdown()

        self.browser.find_element_by_id(f"class_{class_id}").click()

        return self

    def change_game_worksheet(self, worksheet_id):
        self._click_change_worksheet_dropdown()

        self.browser.find_element_by_id(f"worksheet_{worksheet_id}").click()

        self.confirm_dialog()

        return self

    def delete_games(self, game_ids):
        # Tick checkboxes
        for game_id in game_ids:
            self.browser.find_element_by_xpath(
                f"//input[@name='game_ids' and @value='{game_id}']"
            ).click()

        # Click delete
        self.browser.find_element_by_id("deleteGamesButton").click()

        self.confirm_dialog()

        return self

    def _click_change_worksheet_confirm_button(self):
        self.browser.find_element_by_id("confirm_button").click()

    def _click_add_game_dropdown(self):
        self.browser.find_element_by_id("add_class_dropdown").click()

    def _click_change_worksheet_dropdown(self):
        self.browser.find_element_by_id("worksheets_dropdown").click()



import re
from enum import Enum, auto

from django import forms
from django.contrib.auth import update_session_auth_hash


class PasswordStrength(Enum):
    STUDENT = auto()
    INDEPENDENT = auto()
    TEACHER = auto()

    def password_test(self, password):
        if self is PasswordStrength.STUDENT:
            minimum_password_length = 6
            # Make student password case insensitive
            password = password.lower()
            if password and not password_strength_test(
                password=password,
                minimum_password_length=minimum_password_length,
                upper=False,
                lower=False,
                numbers=False,
                special_char=False,
            ):
                raise forms.ValidationError(
                    f"Password not strong enough, consider using at least {minimum_password_length} characters and making it hard to guess."
                )
        elif self is PasswordStrength.INDEPENDENT:
            minimum_password_length = 8
            if password and not password_strength_test(
                password=password,
                minimum_password_length=minimum_password_length,
                upper=True,
                lower=True,
                numbers=True,
                special_char=False,
            ):
                raise forms.ValidationError(
                    f"Password not strong enough, consider using at least {minimum_password_length} characters, "
                    "upper and lower case letters, and numbers and making it hard to guess."
                )
        else:
            minimum_password_length = 10
            if password and not password_strength_test(
                password=password,
                minimum_password_length=minimum_password_length,
                upper=True,
                lower=True,
                numbers=True,
                special_char=True,
            ):
                raise forms.ValidationError(
                    f"Password not strong enough, consider using at least {minimum_password_length} characters, "
                    "upper and lower case letters, numbers, special characters and making it hard to guess."
                )
        return password


def password_strength_test(
    password,
    minimum_password_length,
    upper=True,
    lower=True,
    numbers=True,
    special_char=True,
):
    most_used_passwords = [
        "Abcd1234",
        "Password1",
        "Qwerty123",
        "password",
        "qwerty",
        "abcdef",
    ]
    return (
        len(password) >= minimum_password_length
        and (not upper or re.search(r"[A-Z]", password))
        and (not lower or re.search(r"[a-z]", password))
        and (not numbers or re.search(r"[0-9]", password))
        and (
            not special_char
            or re.search(r"[!@#$%^&*()_+\-=\[\]{};':\"\\|,.<>\/?]", password)
        )
        and (password not in most_used_passwords)
    )


def form_clean_password(self, password_field_name, strength: PasswordStrength):
    password = self.cleaned_data.get(password_field_name, None)
    password = strength.password_test(password)
    return password


def check_update_password(form, user, request, data):
    changing_password = False
    if data["password"] != "":
        changing_password = True
        user.set_password(data["password"])
        user.save()
        update_session_auth_hash(request, form.user)

    return changing_password



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_banner 1'] = '''<div class="banner banner--teacher">
    <div class="container">
        <div class="row">
            <div class="col-sm-12 d-flex">
                <div>
                    <h1 class="banner__text--primary">Test title</h1>
                    
                        <h4>Test subtitle</h4>
                    
                    
                        <p>Test text</p>
                    
                    
                </div>
                <div>
                    <div class="banner--picture">
                        <div class="banner--picture__inside1">
                            <div class="banner--picture__inside2 test--image--class"
                                 
                                 >
                            </div>
                        </div>
                    </div>
                </div>
            </div>
        </div>
    </div>
</div>
'''

snapshots['test_benefits 1'] = '''

<div class="grid-benefits col-sm-8 col-center">
    
    
    
    
        <h5 class="grid-benefits__title grid-benefits__title1">Test title</h5>
    
    
        <h5 class="grid-benefits__title grid-benefits__title2">Test title</h5>
    
    
        <h5 class="grid-benefits__title grid-benefits__title3">Test title</h5>
    
    <p class="grid-benefits__text1">Test text</p>
    <p class="grid-benefits__text2">Test text</p>
    <p class="grid-benefits__text3">Test text</p>
    
        <div class="grid-benefits__button grid-benefits__button1">
        
            <a href="/" class="button button--secondary button--secondary--dark">Test button</a>
        
        </div>
    
    
        <div class="grid-benefits__button grid-benefits__button2">
        
            <a href="/" class="button button--secondary button--secondary--dark">Test button</a>
        
        </div>
    
    
        <div class="grid-benefits__button grid-benefits__button3">
        
            <a href="/" class="button button--secondary button--secondary--dark">Test button</a>
        
        </div>
    
</div>
'''

snapshots['test_card_list 1'] = '''


<script type="text/javascript" src="/static/portal/js/carouselCards.js"></script>

<div id="carouselCards" class="carousel carousel-cards slide" data-ride="carousel" data-interval="false">
    <div class="carousel-inner" role="listbox">
        
            <div class="item col-xs-4 active">
                <div class="card">
                    <div class="card__images">
                        <img class="card__image" src="https://storage.googleapis.com/codeforlife-assets/images/worksheets/future2.jpg">
                    </div>
                    <div class="card__text">
                        <h5 class="card__title">Test card 1</h5>
                        
                            <p>Test description 1</p>
                        
                    </div>
                </div>
            </div>
        
            <div class="item col-xs-4 ">
                <div class="card">
                    <div class="card__images">
                        <img class="card__image" src="https://storage.googleapis.com/codeforlife-assets/images/worksheets/ancient.jpg">
                    </div>
                    <div class="card__text">
                        <h5 class="card__title">Test card 2</h5>
                        
                            <p>Test description 2</p>
                        
                    </div>
                </div>
            </div>
        
            <div class="item col-xs-4 ">
                <div class="card">
                    <div class="card__images">
                        <img class="card__image" src="https://storage.googleapis.com/codeforlife-assets/images/worksheets/modern_day.jpg">
                    </div>
                    <div class="card__text">
                        <h5 class="card__title">Test card 3</h5>
                        
                            <p>Test description 3</p>
                        
                    </div>
                </div>
            </div>
        
            <div class="item col-xs-4 ">
                <div class="card">
                    <div class="card__images">
                        <img class="card__image" src="https://storage.googleapis.com/codeforlife-assets/images/worksheets/prehistory.jpg">
                    </div>
                    <div class="card__text">
                        <h5 class="card__title">Test card 4</h5>
                        
                            <p>Test description 4</p>
                        
                    </div>
                </div>
            </div>
        
            <div class="item col-xs-4 ">
                <div class="card">
                    <div class="card__images">
                        <img class="card__image" src="https://storage.googleapis.com/codeforlife-assets/images/worksheets/broken_future.jpg">
                    </div>
                    <div class="card__text">
                        <h5 class="card__title">Test card 5</h5>
                        
                            <p>Test description 5</p>
                        
                    </div>
                </div>
            </div>
        
    </div>

    <div class="carousel-nav text-center">
        <a class="button prev disabled" data-target="#carouselCards" data-slide="prev">
            <span class="iconify" data-icon="mdi:chevron-left" aria-hidden="true"></span>
            <span class="sr-only">Previous</span>
        </a>
        <a class="button next" data-target="#carouselCards" data-slide="next">
            <span class="iconify" data-icon="mdi:chevron-right" aria-hidden="true"></span>
            <span class="sr-only">Next</span>
        </a>
    </div>
</div>

<script>
    setUpCarouselCards(5)
</script>
'''

snapshots['test_character_list 1'] = '''


<div class="grid grid-characters grid__fit">
    
        <div>
            <img class="grid-character-image" src="https://storage.googleapis.com/codeforlife-assets/images/aimmo_characters/Xian.png" alt="Illustration of Xian">
            <h5 class="grid-character-title">Xian</h5>
            <p>Fun, active, will dance to just about anything that produces a beat. Has great memory, always a joke at hand, might try to introduce memes in Ancient Greece. Scored gold in a track race once and will take any opportunity to bring that up.</p>
        </div>
    
        <div>
            <img class="grid-character-image" src="https://storage.googleapis.com/codeforlife-assets/images/aimmo_characters/Jools.png" alt="Illustration of Jools">
            <h5 class="grid-character-title">Jools</h5>
            <p>A quick-witted kid who wasn&#x27;t expecting to embark in a time-warping journey but can&#x27;t say no to a challenge. Someone has to keep the rest of the group in check, after all!</p>
        </div>
    
        <div>
            <img class="grid-character-image" src="https://storage.googleapis.com/codeforlife-assets/images/aimmo_characters/Zayed.png" alt="Illustration of Zayed">
            <h5 class="grid-character-title">Zayed</h5>
            <p>A pretty chill, curious soul that prefers practice to theory. Always ready to jump into an adventure if it looks interesting enough; not so much otherwise. Probably the one who accidentally turned the time machine on in the first place.</p>
        </div>
    
</div>
'''

snapshots['test_headline 1'] = '''<section>
    <h4>Test title</h4>
</section>
<p class="container">Test description</p>
'''

snapshots['test_hero_card 1'] = '''


<div class="card activated-card">
    <div class="activated-text">Activated</div>
    <img class="card__image" src="https://storage.googleapis.com/codeforlife-assets/images/worksheets/future_active.png">
    <div class="card__text">
        <h4 class="card__title">Test title</h4>
        <p>Test description</p>
        <div class="button-group button-group__icon">
            <a target="_blank" href="https://www.codeforlife.education" class="button button--secondary button--secondary--dark button--icon">
                Test button 1<span class="iconify" data-icon="mdi:open-in-new"></span>
            </a>
            <a href="/kurono/play/1/" class="button button--primary button-right-arrow">
                Test button 2
            </a>
        </div>
    </div>
</div>
'''



from django.views.decorators.cache import never_cache
from two_factor.views.profile import DisableView
from two_factor.views.utils import class_view_decorator

from .form import DisableForm


# This is not changed but imports the from form.py so it overwrites the disable 2FA form
@class_view_decorator(never_cache)
class CustomDisableView(DisableView):
    form_class = DisableForm



from typing import Any, Dict, List, Optional

from aimmo.game_creator import create_game
from aimmo.models import Game
from aimmo.worksheets import WORKSHEETS, Worksheet, get_worksheets_excluding_id
from common.permissions import logged_in_as_student, logged_in_as_teacher
from common.utils import LoginRequiredNoErrorMixin
from django.contrib import messages
from django.contrib.auth.mixins import UserPassesTestMixin
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import QuerySet
from django.urls import reverse_lazy
from django.views.generic.base import TemplateView
from django.views.generic.edit import CreateView

from portal.forms.add_game import AddGameForm
from portal.strings.student_aimmo_dashboard import AIMMO_DASHBOARD_BANNER


class TeacherAimmoDashboard(LoginRequiredNoErrorMixin, UserPassesTestMixin, CreateView):
    login_url = reverse_lazy("teacher_login")
    form_class = AddGameForm
    template_name = "portal/teach/teacher_aimmo_dashboard.html"

    def test_func(self) -> Optional[bool]:
        return logged_in_as_teacher(self.request.user)

    def get_form(self, form_class=None):
        user = self.request.user
        classes = user.userprofile.teacher.class_teacher.all()
        if form_class is None:
            form_class = self.get_form_class()
        return form_class(classes, **self.get_form_kwargs())

    def form_valid(self, form):
        create_game(self.request.user, form)
        return super().form_valid(form)

    def form_invalid(self, form: AddGameForm):
        messages.warning(
            self.request,
            ", ".join(message for errors in form.errors.values() for message in errors),
        )
        return super().form_invalid(form)

    def get_success_url(self):
        return reverse_lazy("teacher_aimmo_dashboard")


class StudentAimmoDashboard(
    LoginRequiredNoErrorMixin, UserPassesTestMixin, TemplateView
):
    template_name = "portal/play/student_aimmo_dashboard.html"

    login_url = reverse_lazy("student_login_access_code")

    def test_func(self) -> Optional[bool]:
        return logged_in_as_student(self.request.user)

    def get_context_data(self, **kwargs: Any) -> Dict[str, Any]:
        student = self.request.user.new_student
        klass = student.class_field

        if klass is None:
            return {"BANNER": AIMMO_DASHBOARD_BANNER}

        aimmo_game = klass.active_game
        if aimmo_game:
            active_worksheet = WORKSHEETS.get(aimmo_game.worksheet_id)
            inactive_worksheets = get_worksheets_excluding_id(active_worksheet.id)

            return {
                "BANNER": AIMMO_DASHBOARD_BANNER,
                "HERO_CARD": self._get_hero_card(active_worksheet, aimmo_game),
                "CARD_LIST": {"cards": self._get_card_list(inactive_worksheets)},
            }
        else:
            return {"BANNER": AIMMO_DASHBOARD_BANNER}

    def _get_hero_card(
        self, active_worksheet: Worksheet, aimmo_game: Game
    ) -> Dict[str, Any]:
        return {
            "image": active_worksheet.active_image_path,
            "title": active_worksheet.name,
            "description": active_worksheet.description,
            "button1": {
                "text": "Read challenge",
                "url": active_worksheet.student_challenge_url,
            },
            "button2": {
                "text": "Play game",
                "url": "kurono/play",
                "url_args": aimmo_game.id,
            },
        }

    def _get_card_list(self, inactive_worksheets: QuerySet) -> List[Dict[str, Any]]:
        return [
            {
                "image": inactive_worksheet.image_path,
                "title": inactive_worksheet.name,
                "description": inactive_worksheet.short_description,
            }
            for inactive_worksheet in inactive_worksheets
        ]



import pytest
from aimmo.models import Game
from aimmo.worksheets import WORKSHEETS
from common.models import Class
from common.tests.utils.classes import create_class_directly
from common.tests.utils.organisation import (
    create_organisation_directly,
)
from common.tests.utils.student import create_school_student_directly
from common.tests.utils.teacher import signup_teacher_directly
from django.test.client import Client
from django.urls.base import reverse

from .base_test import BaseTest
from .conftest import IndependentStudent, SchoolStudent


@pytest.mark.django_db
def test_student_cannot_access_teacher_dashboard(
    student1: SchoolStudent, class1: Class
):
    """
    Given you are logged in as a student,
    When you try to access the teacher dashboard,
    Then you cannot access it and are instead redirected.
    """
    c = Client()
    url = reverse("student_login", kwargs={"access_code": class1.access_code})
    data = {
        "username": student1.username,
        "password": student1.password,
    }

    c.post(url, data)

    student_dashboard_url = reverse("student_aimmo_dashboard")

    response_s = c.get(student_dashboard_url)

    assert response_s.status_code == 200

    teacher_dashboard_url = reverse("teacher_aimmo_dashboard")

    response = c.get(teacher_dashboard_url)

    assert response.status_code == 302


@pytest.mark.django_db
def test_indep_student_cannot_access_dashboard(
    independent_student1: IndependentStudent,
):
    """
    Given you are logged in as an independent student,
    When you try to access the student dashboard,
    Then you can access it but the context only has the banner.
    """
    c = Client()
    url = reverse("independent_student_login")
    data = {
        "username": independent_student1.username,
        "password": independent_student1.password,
    }

    c.post(url, data)

    student_dashboard_url = reverse("student_aimmo_dashboard")

    response = c.get(student_dashboard_url)

    assert response.status_code == 200
    assert "BANNER" in response.context
    assert "HERO_CARD" not in response.context
    assert "CARD_LIST" not in response.context


@pytest.mark.django_db
def test_student_aimmo_dashboard_loads(
    student1: SchoolStudent, class1: Class, aimmo_game1: Game
):
    """
    Given an aimmo game is linked to a class,
    When a student of that class goes on the Student Kurono Dashboard page,
    Then the page loads and the context contains the hero card and card list
    associated to the aimmo game.

    Then, given that the class no longer has a game linked to it,
    When the student goes on the same page,
    Then the page still loads but the context no longer contains the hero card
    or the card list elements.
    """
    c = Client()
    student_login_url = reverse(
        "student_login", kwargs={"access_code": class1.access_code}
    )
    data = {
        "username": student1.username,
        "password": student1.password,
    }

    c.post(student_login_url, data)

    student_dashboard_url = reverse("student_aimmo_dashboard")
    response = c.get(student_dashboard_url)

    assert response.status_code == 200
    assert "HERO_CARD" in response.context
    assert "CARD_LIST" in response.context

    aimmo_game1.delete()

    url = reverse("student_aimmo_dashboard")
    response = c.get(url)

    assert response.status_code == 200
    assert "HERO_CARD" not in response.context
    assert "CARD_LIST" not in response.context


# Selenium tests
class TestAimmoDashboardFrontend(BaseTest):
    def test_worksheet_dropdown_changes_worksheet(self):
        teacher_email, teacher_password = signup_teacher_directly()
        create_organisation_directly(teacher_email)
        klass, class_name, access_code = create_class_directly(teacher_email)
        student_name, student_password, _ = create_school_student_directly(access_code)

        worksheet1 = WORKSHEETS.get(1)
        worksheet2 = WORKSHEETS.get(2)

        self.selenium.get(self.live_server_url)
        page = (
            self.go_to_homepage()
            .go_to_teacher_login_page()
            .login(teacher_email, teacher_password)
        )
        page = page.go_to_kurono_teacher_dashboard_page().create_game(klass.id)

        game = Game.objects.get(game_class=klass)

        assert game.worksheet == worksheet1

        page.change_game_worksheet(worksheet2.id)

        game = Game.objects.get(game_class=klass)

        assert game.worksheet == worksheet2

    def test_delete_games(self):
        teacher_email, teacher_password = signup_teacher_directly()
        create_organisation_directly(teacher_email)

        klass1, _, _ = create_class_directly(teacher_email)
        game1 = Game(game_class=klass1)
        game1.save()

        klass2, _, _ = create_class_directly(teacher_email)
        game2 = Game(game_class=klass2)
        game2.save()

        assert Game.objects.count() == 2

        self.selenium.get(self.live_server_url)
        page = (
            self.go_to_homepage()
            .go_to_teacher_login_page()
            .login(teacher_email, teacher_password)
        )
        page.go_to_kurono_teacher_dashboard_page().delete_games([game1.id, game2.id])

        assert Game.objects.filter(is_archived=False).count() == 0
        assert Game.objects.filter(is_archived=True).count() == 2
